import logging
import math
from typing import Any, Callable, Dict, List, Optional, Tuple, TypeVar, Union

import ray
from ray.data._internal.execution.interfaces import RefBundle
from ray.data._internal.planner.exchange.interfaces import ExchangeTaskScheduler
from ray.data._internal.progress_bar import ProgressBar
from ray.data._internal.remote_fn import cached_remote_fn
from ray.data._internal.stats import StatsDict
from ray.data.block import Block, BlockAccessor, BlockExecStats, BlockMetadata
from ray.data.context import DatasetContext
from ray.types import ObjectRef
from ray.util.scheduling_strategies import NodeAffinitySchedulingStrategy

logger = logging.getLogger(__name__)


T = TypeVar("T")
U = TypeVar("U")


class _MergeTaskSchedule:
    def __init__(self, output_num_blocks: int, num_merge_tasks_per_round: int):
        self.output_num_blocks = output_num_blocks
        self.num_merge_tasks_per_round = num_merge_tasks_per_round
        self.merge_partition_size = output_num_blocks // num_merge_tasks_per_round
        self._partitions_with_extra_task = output_num_blocks % num_merge_tasks_per_round

    def get_num_reducers_per_merge_idx(self, merge_idx: int) -> int:
        """
        Each intermediate merge task will produce outputs for a partition of P
        final reduce tasks. This helper function returns P based on the merge
        task index.
        """
        assert merge_idx < self.num_merge_tasks_per_round
        partition_size = self.merge_partition_size
        if merge_idx < self._partitions_with_extra_task:
            partition_size += 1
        return partition_size

    def get_merge_idx_for_reducer_idx(self, reducer_idx: int) -> int:
        if reducer_idx < self.merge_partition_size * self._partitions_with_extra_task:
            merge_idx = reducer_idx // (self.merge_partition_size + 1)
        else:
            reducer_idx -= (
                self.merge_partition_size + 1
            ) * self._partitions_with_extra_task
            merge_idx = (
                self._partitions_with_extra_task
                + reducer_idx // self.merge_partition_size
            )
        assert merge_idx < self.num_merge_tasks_per_round
        return merge_idx

    def round_robin_reduce_idx_iterator(self):
        """
        When there are multiple nodes, merge tasks are spread throughout the
        cluster to improve load-balancing. Each merge task produces outputs for
        a contiguous partition of reduce tasks. This method creates an iterator
        that returns reduce task indices round-robin across the merge tasks.
        This can be used to submit reduce tasks in a way that spreads the load
        evenly across the cluster.
        """
        idx = 0
        round_idx = 0
        while idx < self.output_num_blocks:
            for merge_idx in range(self.num_merge_tasks_per_round):
                if merge_idx < self._partitions_with_extra_task:
                    reduce_idx = merge_idx * (self.merge_partition_size + 1)
                    partition_size = self.merge_partition_size + 1
                else:
                    reduce_idx = self._partitions_with_extra_task * (
                        self.merge_partition_size + 1
                    )
                    merge_idx -= self._partitions_with_extra_task
                    reduce_idx += merge_idx * self.merge_partition_size
                    partition_size = self.merge_partition_size

                if round_idx >= partition_size:
                    continue

                reduce_idx += round_idx
                yield reduce_idx
                idx += 1
            round_idx += 1


class _PushBasedShuffleStage:
    def __init__(
        self,
        output_num_blocks: int,
        num_rounds: int,
        num_map_tasks_per_round: int,
        merge_task_placement: List[str],
    ):
        self.num_rounds = num_rounds
        self.num_map_tasks_per_round = num_map_tasks_per_round
        self.num_merge_tasks_per_round = len(merge_task_placement)

        node_strategies = {
            node_id: {
                "scheduling_strategy": NodeAffinitySchedulingStrategy(
                    node_id, soft=True
                )
            }
            for node_id in set(merge_task_placement)
        }
        self._merge_task_options = [
            node_strategies[node_id] for node_id in merge_task_placement
        ]

        self.merge_schedule = _MergeTaskSchedule(
            output_num_blocks, self.num_merge_tasks_per_round
        )

    def get_merge_task_options(self, merge_idx):
        return self._merge_task_options[merge_idx]


class _PipelinedStageExecutor:
    def __init__(
        self,
        stage_iter,
        num_tasks_per_round: int,
        max_concurrent_rounds: int = 1,
        progress_bar: Optional[ProgressBar] = None,
    ):
        self._stage_iter = stage_iter
        self._num_tasks_per_round = num_tasks_per_round
        self._max_concurrent_rounds = max_concurrent_rounds
        self._progress_bar = progress_bar

        self._rounds: List[List[ObjectRef]] = []
        self._task_idx = 0

        self._submit_round()

    def __iter__(self):
        return self

    def __next__(self):
        """
        Submit one round of tasks. If we already have the max concurrent rounds
        in flight, first wait for the oldest round of tasks to finish.
        """
        prev_metadata = []
        if all(len(r) == 0 for r in self._rounds):
            raise StopIteration

        if len(self._rounds) >= self._max_concurrent_rounds:
            prev_metadata_refs = self._rounds.pop(0)
            if prev_metadata_refs:
                if self._progress_bar is not None:
                    prev_metadata = self._progress_bar.fetch_until_complete(
                        prev_metadata_refs
                    )
                else:
                    prev_metadata = ray.get(prev_metadata_refs)

        self._submit_round()

        return prev_metadata

    def _submit_round(self):
        assert len(self._rounds) < self._max_concurrent_rounds
        task_round = []
        for _ in range(self._num_tasks_per_round):
            try:
                task_round.append(next(self._stage_iter))
            except StopIteration:
                break
        self._rounds.append(task_round)


class _MapStageIterator:
    def __init__(self, input_blocks_list, shuffle_map, map_args):
        self._input_blocks_list = input_blocks_list
        self._shuffle_map = shuffle_map
        self._map_args = map_args

        self._mapper_idx = 0
        self._map_results = []

    def __iter__(self):
        return self

    def __next__(self):
        if not self._input_blocks_list:
            raise StopIteration

        block = self._input_blocks_list.pop(0)
        # NOTE(swang): Results are shuffled between map and merge tasks, so
        # there is no advantage to colocating specific map and merge tasks.
        # Therefore, we do not specify a node affinity policy for map tasks
        # in case the caller or Ray has a better scheduling strategy, e.g.,
        # based on data locality.
        map_result = self._shuffle_map.remote(
            self._mapper_idx,
            block,
            *self._map_args,
        )
        metadata_ref = map_result.pop(-1)
        self._map_results.append(map_result)
        self._mapper_idx += 1
        return metadata_ref

    def pop_map_results(self) -> List[List[ObjectRef]]:
        map_results = self._map_results
        self._map_results = []
        return map_results


class _MergeStageIterator:
    def __init__(
        self,
        map_stage_iter: _MapStageIterator,
        shuffle_merge,
        stage: _PushBasedShuffleStage,
        reduce_args,
    ):
        self._map_stage_iter = map_stage_iter
        self._shuffle_merge = shuffle_merge
        self._stage = stage
        self._reduce_args = reduce_args

        self._merge_idx = 0
        self._map_result_buffer = None
        # Final outputs from the map-merge stage.
        # This is a map from merge task index to a nested list of merge results
        # (ObjectRefs). Each merge task index corresponds to a partition of P
        # final reduce tasks.
        self._all_merge_results = [
            [] for _ in range(self._stage.num_merge_tasks_per_round)
        ]

    def __next__(self):
        if not self._map_result_buffer or not self._map_result_buffer[0]:
            assert self._merge_idx == 0
            self._map_result_buffer = self._map_stage_iter.pop_map_results()

        if not self._map_result_buffer:
            raise StopIteration

        # Shuffle the map results for the merge tasks.
        merge_args = [map_result.pop(0) for map_result in self._map_result_buffer]
        num_merge_returns = self._stage.merge_schedule.get_num_reducers_per_merge_idx(
            self._merge_idx
        )
        merge_result = self._shuffle_merge.options(
            num_returns=1 + num_merge_returns,
            **self._stage.get_merge_task_options(self._merge_idx),
        ).remote(
            *merge_args,
            reduce_args=self._reduce_args,
        )
        metadata_ref = merge_result.pop(-1)
        self._all_merge_results[self._merge_idx].append(merge_result)
        del merge_result

        self._merge_idx += 1
        self._merge_idx %= self._stage.num_merge_tasks_per_round
        return metadata_ref

    def pop_merge_results(self) -> List[List[ObjectRef]]:
        all_merge_results = self._all_merge_results
        self._all_merge_results = []
        return all_merge_results


class _ReduceStageIterator:
    def __init__(
        self,
        stage: _PushBasedShuffleStage,
        shuffle_reduce,
        all_merge_results: List[List[List[ObjectRef]]],
        ray_remote_args,
        reduce_args: List[Any],
    ):
        self._shuffle_reduce = shuffle_reduce
        self._stage = stage
        self._reduce_arg_blocks: List[Tuple[int, List[ObjectRef]]] = []
        self._ray_remote_args = ray_remote_args
        self._reduce_args = reduce_args

        for reduce_idx in self._stage.merge_schedule.round_robin_reduce_idx_iterator():
            merge_idx = self._stage.merge_schedule.get_merge_idx_for_reducer_idx(
                reduce_idx
            )
            reduce_arg_blocks = [
                merge_results.pop(0) for merge_results in all_merge_results[merge_idx]
            ]
            self._reduce_arg_blocks.append((reduce_idx, reduce_arg_blocks))
        assert len(self._reduce_arg_blocks) == stage.merge_schedule.output_num_blocks

        for merge_idx, merge_results in enumerate(all_merge_results):
            assert all(len(merge_result) == 0 for merge_result in merge_results), (
                "Reduce stage did not process outputs from merge tasks at index: "
                f"{merge_idx}"
            )

        self._reduce_results: List[Tuple[int, ObjectRef]] = []

    def __iter__(self):
        return self

    def __next__(self):
        if not self._reduce_arg_blocks:
            raise StopIteration

        reduce_idx, reduce_arg_blocks = self._reduce_arg_blocks.pop(0)
        merge_idx = self._stage.merge_schedule.get_merge_idx_for_reducer_idx(reduce_idx)
        # Submit one partition of reduce tasks, one for each of the P
        # outputs produced by the corresponding merge task.
        # We also add the merge task arguments so that the reduce task
        # is colocated with its inputs.
        block, meta = self._shuffle_reduce.options(
            **self._ray_remote_args,
            **self._stage.get_merge_task_options(merge_idx),
            num_returns=2,
        ).remote(*self._reduce_args, *reduce_arg_blocks, partial_reduce=False)
        self._reduce_results.append((reduce_idx, block))
        return meta

    def pop_reduce_results(self):
        reduce_results = self._reduce_results
        self._reduce_results = []
        return reduce_results


class PushBasedShuffleTaskScheduler(ExchangeTaskScheduler):
    """
    Push-based shuffle merges intermediate map outputs on the reducer nodes
    while other map tasks are executing. The merged outputs are merged again
    during a final reduce stage. This works as follows:

    1. Submit rounds of concurrent map and merge tasks until all map inputs
    have been processed. In each round, we execute:

       M map tasks
         Each produces N outputs. Each output contains P blocks.
       N merge tasks
         Takes 1 output from each of M map tasks.
         Each produces P outputs.
       Where M and N are chosen to maximize parallelism across CPUs. Note that
       this assumes that all CPUs in the cluster will be dedicated to the
       shuffle job.

       Map and merge tasks are pipelined so that we always merge the previous
       round of map outputs while executing the next round of map tasks.

    2. In the final reduce stage:
       R reduce tasks
         Takes 1 output from one of the merge tasks from every round.

    Notes:
        N * P = R = total number of output blocks
        M / N = merge factor - the ratio of map : merge tasks is to improve
          pipelined parallelism. For example, if map takes twice as long to
          execute as merge, then we should set this to 2.
        See paper at https://arxiv.org/abs/2203.05072 for more details.
    """

    def execute(
        self,
        refs: List[RefBundle],
        output_num_blocks: int,
        map_ray_remote_args: Optional[Dict[str, Any]] = None,
        reduce_ray_remote_args: Optional[Dict[str, Any]] = None,
        merge_factor: int = 2,
    ) -> Tuple[List[RefBundle], StatsDict]:
        logger.info("Using experimental push-based shuffle.")
        # TODO: Preemptively clear the blocks list since we will incrementally delete
        # the last remaining references as we submit the dependent map tasks during the
        # map-merge stage.

        # TODO(swang): For jobs whose reduce work is heavier than the map work,
        # we should support fractional merge factors.
        # TODO(swang): For large jobs, we should try to choose the merge factor
        # automatically, e.g., by running one test round of map and merge tasks
        # and comparing their run times.
        # TODO(swang): Add option to automatically reduce write amplification
        # during map-merge stage, by limiting how many partitions can be
        # processed concurrently.
        input_blocks_list = []
        for ref_bundle in refs:
            for block, _ in ref_bundle.blocks:
                input_blocks_list.append(block)
        input_owned = all(b.owns_blocks for b in refs)

        if map_ray_remote_args is None:
            map_ray_remote_args = {}
        if reduce_ray_remote_args is None:
            reduce_ray_remote_args = {}
        # The placement strategy for reduce tasks is overwritten to colocate
        # them with their inputs from the merge stage, so remove any
        # pre-specified scheduling strategy here.
        reduce_ray_remote_args = reduce_ray_remote_args.copy()
        reduce_ray_remote_args.pop("scheduling_strategy", None)

        # Compute all constants used for task scheduling.
        num_cpus_per_node_map = _get_num_cpus_per_node_map()
        stage = self._compute_shuffle_schedule(
            num_cpus_per_node_map,
            len(input_blocks_list),
            merge_factor,
            output_num_blocks,
        )

        map_fn = self._map_partition
        merge_fn = self._merge

        def map_partition(*args, **kwargs):
            return map_fn(self._exchange_spec.map, *args, **kwargs)

        def merge(*args, **kwargs):
            return merge_fn(self._exchange_spec.reduce, *args, **kwargs)

        shuffle_map = cached_remote_fn(map_partition)
        shuffle_map = shuffle_map.options(
            **map_ray_remote_args,
            num_returns=1 + stage.num_merge_tasks_per_round,
        )

        map_stage_iter = _MapStageIterator(
            input_blocks_list,
            shuffle_map,
            [output_num_blocks, stage.merge_schedule, *self._exchange_spec._map_args],
        )
        map_bar = ProgressBar("Shuffle Map", position=0, total=len(input_blocks_list))
        map_stage_executor = _PipelinedStageExecutor(
            map_stage_iter, stage.num_map_tasks_per_round, progress_bar=map_bar
        )

        shuffle_merge = cached_remote_fn(merge)
        merge_stage_iter = _MergeStageIterator(
            map_stage_iter, shuffle_merge, stage, self._exchange_spec._reduce_args
        )
        merge_stage_executor = _PipelinedStageExecutor(
            merge_stage_iter, stage.num_merge_tasks_per_round, max_concurrent_rounds=2
        )

        # Execute the map-merge stage. This submits tasks in rounds of M map
        # tasks and N merge tasks each. Task execution between map and merge is
        # pipelined, so that while executing merge for one round of inputs, we
        # also execute the map tasks for the following round.
        map_done = False
        merge_done = False
        map_stage_metadata = []
        merge_stage_metadata = []
        while not (map_done and merge_done):
            try:
                map_stage_metadata += next(map_stage_executor)
            except StopIteration:
                map_done = True
                break

            try:
                merge_stage_metadata += next(merge_stage_executor)
            except StopIteration:
                merge_done = True
                break

        map_bar.close()
        all_merge_results = merge_stage_iter.pop_merge_results()

        # Execute and wait for the reduce stage.
        reduce_bar = ProgressBar("Shuffle Reduce", total=output_num_blocks)
        shuffle_reduce = cached_remote_fn(self._exchange_spec.reduce)
        reduce_stage_iter = _ReduceStageIterator(
            stage,
            shuffle_reduce,
            all_merge_results,
            reduce_ray_remote_args,
            self._exchange_spec._reduce_args,
        )

        max_reduce_tasks_in_flight = output_num_blocks
        ctx = DatasetContext.get_current()
        if ctx.pipeline_push_based_shuffle_reduce_tasks:
            # If pipelining is enabled, we should still try to utilize all
            # cores.
            max_reduce_tasks_in_flight = min(
                max_reduce_tasks_in_flight, sum(num_cpus_per_node_map.values())
            )

        reduce_stage_executor = _PipelinedStageExecutor(
            reduce_stage_iter,
            max_reduce_tasks_in_flight,
            max_concurrent_rounds=2,
            progress_bar=reduce_bar,
        )
        reduce_stage_metadata = []
        while True:
            try:
                reduce_stage_metadata += next(reduce_stage_executor)
            except StopIteration:
                break

        new_blocks = reduce_stage_iter.pop_reduce_results()
        sorted_blocks = [
            (block[0], block[1], reduce_stage_metadata[i])
            for i, block in enumerate(new_blocks)
        ]
        sorted_blocks.sort(key=lambda x: x[0])
        _, new_blocks, reduce_stage_metadata = zip(*sorted_blocks)
        del sorted_blocks

        assert (
            len(new_blocks) == output_num_blocks
        ), f"Expected {output_num_blocks} outputs, produced {len(new_blocks)}"
        reduce_bar.close()

        output = []
        for block, meta in zip(new_blocks, reduce_stage_metadata):
            output.append(
                RefBundle(
                    [
                        (
                            block,
                            meta,
                        )
                    ],
                    owns_blocks=input_owned,
                )
            )
        stats = {
            "map": map_stage_metadata,
            "merge": merge_stage_metadata,
            "reduce": reduce_stage_metadata,
        }

        return (output, stats)

    @staticmethod
    def _map_partition(
        map_fn,
        idx: int,
        block: Block,
        output_num_blocks: int,
        schedule: _MergeTaskSchedule,
        *map_args: List[Any],
    ) -> List[Union[BlockMetadata, Block]]:
        mapper_outputs = map_fn(idx, block, output_num_blocks, *map_args)
        meta = mapper_outputs.pop(-1)

        parts = []
        merge_idx = 0
        while mapper_outputs:
            partition_size = schedule.get_num_reducers_per_merge_idx(merge_idx)
            parts.append(mapper_outputs[:partition_size])
            mapper_outputs = mapper_outputs[partition_size:]
            merge_idx += 1
        assert len(parts) == schedule.num_merge_tasks_per_round, (
            len(parts),
            schedule.num_merge_tasks_per_round,
        )
        return parts + [meta]

    @staticmethod
    def _merge(
        reduce_fn,
        *all_mapper_outputs: List[List[Block]],
        reduce_args: Optional[List[Any]] = None,
    ) -> List[Union[BlockMetadata, Block]]:
        """
        Returns list of [BlockMetadata, O1, O2, O3, ...output_num_blocks].
        """
        assert (
            len({len(mapper_outputs) for mapper_outputs in all_mapper_outputs}) == 1
        ), "Received different number of map inputs"
        stats = BlockExecStats.builder()
        if not reduce_args:
            reduce_args = []

        num_rows = 0
        size_bytes = 0
        schema = None
        for i, mapper_outputs in enumerate(zip(*all_mapper_outputs)):
            block, meta = reduce_fn(*reduce_args, *mapper_outputs, partial_reduce=True)
            yield block

            block = BlockAccessor.for_block(block)
            num_rows += block.num_rows()
            size_bytes += block.size_bytes()
            schema = block.schema()
            del block

        yield BlockMetadata(
            num_rows=num_rows,
            size_bytes=size_bytes,
            schema=schema,
            input_files=None,
            exec_stats=stats.build(),
        )

    @staticmethod
    def _compute_shuffle_schedule(
        num_cpus_per_node_map: Dict[str, int],
        num_input_blocks: int,
        merge_factor: int,
        num_output_blocks: int,
    ) -> _PushBasedShuffleStage:
        num_cpus_total = sum(v for v in num_cpus_per_node_map.values())
        task_parallelism = min(num_cpus_total, num_input_blocks)

        num_tasks_per_map_merge_group = merge_factor + 1
        num_merge_tasks_per_round = 0
        merge_task_placement = []
        leftover_cpus = 0
        # Compute the total number of merge tasks and their node placement.
        # Each merge task should be grouped with `merge_factor` map tasks for
        # pipelining. These groups should then be spread across nodes according
        # to CPU availability for load-balancing.
        for node, num_cpus in num_cpus_per_node_map.items():
            node_parallelism = min(
                num_cpus, num_input_blocks // len(num_cpus_per_node_map)
            )
            num_merge_tasks = node_parallelism // num_tasks_per_map_merge_group
            for i in range(num_merge_tasks):
                merge_task_placement.append(node)
            num_merge_tasks_per_round += num_merge_tasks

            # Handle the case where a single node cannot fit a group of map and
            # merge tasks, but we can spread the group across multiple distinct
            # nodes.
            leftover_cpus += node_parallelism % num_tasks_per_map_merge_group
            if num_merge_tasks == 0 and leftover_cpus > num_tasks_per_map_merge_group:
                merge_task_placement.append(node)
                num_merge_tasks_per_round += 1
                leftover_cpus -= num_tasks_per_map_merge_group
        if num_merge_tasks_per_round == 0:
            merge_task_placement.append(list(num_cpus_per_node_map)[0])
            num_merge_tasks_per_round = 1

        assert num_merge_tasks_per_round == len(merge_task_placement)
        num_map_tasks_per_round = max(task_parallelism - num_merge_tasks_per_round, 1)

        num_rounds = math.ceil(num_input_blocks / num_map_tasks_per_round)
        return _PushBasedShuffleStage(
            num_output_blocks,
            num_rounds,
            num_map_tasks_per_round,
            merge_task_placement,
        )


def _execute_pipelined_stage(
    stage_fn: Callable[[T], Tuple[ObjectRef, U]],
    prev_metadata_refs: List[ObjectRef],
    stage_args: List[T],
    progress_bar: Optional[ProgressBar] = None,
) -> Tuple[List[BlockMetadata], List[ObjectRef], List[U]]:
    """
    Helper function to execute a stage of tasks. This will wait for the
    previous round of tasks to complete before submitting the next.
    """
    # TODO(swang): Straggler tasks can cause pipeline bubbles. Instead of
    # waiting for all previous tasks, we should wait for some tasks on each
    # node to finish.
    if progress_bar is not None:
        prev_metadata = progress_bar.fetch_until_complete(prev_metadata_refs)
    else:
        prev_metadata = ray.get(prev_metadata_refs)
    prev_metadata_refs.clear()

    metadata_refs = []
    data_outputs = []
    while stage_args:
        arg = stage_args.pop(0)
        metadata_ref, data_output = stage_fn(arg)
        metadata_refs.append(metadata_ref)
        data_outputs.append(data_output)
    return prev_metadata, metadata_refs, data_outputs


def _get_num_cpus_per_node_map() -> Dict[str, int]:
    nodes = ray.nodes()
    # Map from per-node resource name to number of CPUs available on that
    # node.
    num_cpus_per_node_map = {}
    for node in nodes:
        resources = node["Resources"]
        num_cpus = int(resources.get("CPU", 0))
        if num_cpus == 0:
            continue
        num_cpus_per_node_map[node["NodeID"]] = num_cpus
    return num_cpus_per_node_map



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: src/ray/protobuf/common.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import runtime_env_common_pb2 as src_dot_ray_dot_protobuf_dot_runtime__env__common__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x1dsrc/ray/protobuf/common.proto\x12\x07ray.rpc\x1a)src/ray/protobuf/runtime_env_common.proto\"M\n\x1eNodeAffinitySchedulingStrategy\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12\x12\n\x04soft\x18\x02 \x01(\x08R\x04soft\"\xdf\x01\n PlacementGroupSchedulingStrategy\x12,\n\x12placement_group_id\x18\x01 \x01(\x0cR\x10placementGroupId\x12?\n\x1cplacement_group_bundle_index\x18\x02 \x01(\x03R\x19placementGroupBundleIndex\x12L\n#placement_group_capture_child_tasks\x18\x03 \x01(\x08R\x1fplacementGroupCaptureChildTasks\"\x1b\n\x19\x44\x65\x66\x61ultSchedulingStrategy\"\x1a\n\x18SpreadSchedulingStrategy\"\xe6\x03\n\x12SchedulingStrategy\x12\x64\n\x1b\x64\x65\x66\x61ult_scheduling_strategy\x18\x01 \x01(\x0b\x32\".ray.rpc.DefaultSchedulingStrategyH\x00R\x19\x64\x65\x66\x61ultSchedulingStrategy\x12z\n#placement_group_scheduling_strategy\x18\x02 \x01(\x0b\x32).ray.rpc.PlacementGroupSchedulingStrategyH\x00R placementGroupSchedulingStrategy\x12\x61\n\x1aspread_scheduling_strategy\x18\x03 \x01(\x0b\x32!.ray.rpc.SpreadSchedulingStrategyH\x00R\x18spreadSchedulingStrategy\x12t\n!node_affinity_scheduling_strategy\x18\x04 \x01(\x0b\x32\'.ray.rpc.NodeAffinitySchedulingStrategyH\x00R\x1enodeAffinitySchedulingStrategyB\x15\n\x13scheduling_strategy\"v\n\x07\x41\x64\x64ress\x12\x1b\n\traylet_id\x18\x01 \x01(\x0cR\x08rayletId\x12\x1d\n\nip_address\x18\x02 \x01(\tR\tipAddress\x12\x12\n\x04port\x18\x03 \x01(\x05R\x04port\x12\x1b\n\tworker_id\x18\x04 \x01(\x0cR\x08workerId\"z\n\x16JavaFunctionDescriptor\x12\x1d\n\nclass_name\x18\x01 \x01(\tR\tclassName\x12#\n\rfunction_name\x18\x02 \x01(\tR\x0c\x66unctionName\x12\x1c\n\tsignature\x18\x03 \x01(\tR\tsignature\"\xa4\x01\n\x18PythonFunctionDescriptor\x12\x1f\n\x0bmodule_name\x18\x01 \x01(\tR\nmoduleName\x12\x1d\n\nclass_name\x18\x02 \x01(\tR\tclassName\x12#\n\rfunction_name\x18\x03 \x01(\tR\x0c\x66unctionName\x12#\n\rfunction_hash\x18\x04 \x01(\tR\x0c\x66unctionHash\"s\n\x15\x43ppFunctionDescriptor\x12#\n\rfunction_name\x18\x01 \x01(\tR\x0c\x66unctionName\x12\x16\n\x06\x63\x61ller\x18\x02 \x01(\tR\x06\x63\x61ller\x12\x1d\n\nclass_name\x18\x03 \x01(\tR\tclassName\"\xc5\x02\n\x12\x46unctionDescriptor\x12[\n\x18java_function_descriptor\x18\x01 \x01(\x0b\x32\x1f.ray.rpc.JavaFunctionDescriptorH\x00R\x16javaFunctionDescriptor\x12\x61\n\x1apython_function_descriptor\x18\x02 \x01(\x0b\x32!.ray.rpc.PythonFunctionDescriptorH\x00R\x18pythonFunctionDescriptor\x12X\n\x17\x63pp_function_descriptor\x18\x03 \x01(\x0b\x32\x1e.ray.rpc.CppFunctionDescriptorH\x00R\x15\x63ppFunctionDescriptorB\x15\n\x13\x66unction_descriptor\"\x9f\x01\n\x10\x43oncurrencyGroup\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\'\n\x0fmax_concurrency\x18\x02 \x01(\x05R\x0emaxConcurrency\x12N\n\x14\x66unction_descriptors\x18\x03 \x03(\x0b\x32\x1b.ray.rpc.FunctionDescriptorR\x13\x66unctionDescriptors\"\xb3\x02\n\x0cRayErrorInfo\x12\x44\n\x10\x61\x63tor_died_error\x18\x02 \x01(\x0b\x32\x18.ray.rpc.ActorDeathCauseH\x00R\x0e\x61\x63torDiedError\x12\x66\n\x1eruntime_env_setup_failed_error\x18\x03 \x01(\x0b\x32 .ray.rpc.RuntimeEnvFailedContextH\x00R\x1aruntimeEnvSetupFailedError\x12%\n\rerror_message\x18\x04 \x01(\tH\x00R\x0c\x65rrorMessage\x12\x36\n\nerror_type\x18\x0b \x01(\x0e\x32\x12.ray.rpc.ErrorTypeH\x01R\terrorType\x88\x01\x01\x42\x07\n\x05\x65rrorB\r\n\x0b_error_type\"\xb0\x02\n\x17OutOfMemoryErrorContext\x12\x17\n\x07task_id\x18\x01 \x01(\x0cR\x06taskId\x12\x1b\n\ttask_name\x18\x02 \x01(\tR\x08taskName\x12&\n\x0fnode_ip_address\x18\x03 \x01(\tR\rnodeIpAddress\x12*\n\x11memory_used_bytes\x18\x04 \x01(\x04R\x0fmemoryUsedBytes\x12,\n\x12memory_total_bytes\x18\x05 \x01(\x04R\x10memoryTotalBytes\x12\x32\n\x15memory_usage_fraction\x18\x06 \x01(\x01R\x13memoryUsageFraction\x12)\n\x10memory_threshold\x18\x07 \x01(\x01R\x0fmemoryThreshold\"W\n\x14NodeDiedErrorContext\x12\x17\n\x07node_id\x18\x01 \x01(\x0cR\x06nodeId\x12&\n\x0fnode_ip_address\x18\x02 \x01(\tR\rnodeIpAddress\"\xae\x01\n\x0cRayException\x12-\n\x08language\x18\x01 \x01(\x0e\x32\x11.ray.rpc.LanguageR\x08language\x12\x31\n\x14serialized_exception\x18\x02 \x01(\x0cR\x13serializedException\x12<\n\x1a\x66ormatted_exception_string\x18\x03 \x01(\tR\x18\x66ormattedExceptionString\"\xd2\x03\n\x0f\x41\x63torDeathCause\x12Z\n\x1d\x63reation_task_failure_context\x18\x01 \x01(\x0b\x32\x15.ray.rpc.RayExceptionH\x00R\x1a\x63reationTaskFailureContext\x12_\n\x1aruntime_env_failed_context\x18\x02 \x01(\x0b\x32 .ray.rpc.RuntimeEnvFailedContextH\x00R\x17runtimeEnvFailedContext\x12Y\n\x18\x61\x63tor_died_error_context\x18\x03 \x01(\x0b\x32\x1e.ray.rpc.ActorDiedErrorContextH\x00R\x15\x61\x63torDiedErrorContext\x12\x64\n\x1b\x61\x63tor_unschedulable_context\x18\x04 \x01(\x0b\x32\".ray.rpc.ActorUnschedulableContextH\x00R\x19\x61\x63torUnschedulableContext\x12\x36\n\x0boom_context\x18\x05 \x01(\x0b\x32\x13.ray.rpc.OomContextH\x00R\noomContextB\t\n\x07\x63ontext\">\n\x17RuntimeEnvFailedContext\x12#\n\rerror_message\x18\x01 \x01(\tR\x0c\x65rrorMessage\"@\n\x19\x41\x63torUnschedulableContext\x12#\n\rerror_message\x18\x01 \x01(\tR\x0c\x65rrorMessage\"\xd3\x02\n\x15\x41\x63torDiedErrorContext\x12#\n\rerror_message\x18\x01 \x01(\tR\x0c\x65rrorMessage\x12\x19\n\x08owner_id\x18\x02 \x01(\x0cR\x07ownerId\x12(\n\x10owner_ip_address\x18\x03 \x01(\tR\x0eownerIpAddress\x12&\n\x0fnode_ip_address\x18\x04 \x01(\tR\rnodeIpAddress\x12\x10\n\x03pid\x18\x05 \x01(\rR\x03pid\x12\x12\n\x04name\x18\x06 \x01(\tR\x04name\x12#\n\rray_namespace\x18\x07 \x01(\tR\x0crayNamespace\x12\x1d\n\nclass_name\x18\x08 \x01(\tR\tclassName\x12\x19\n\x08\x61\x63tor_id\x18\t \x01(\x0cR\x07\x61\x63torId\x12#\n\rnever_started\x18\n \x01(\x08R\x0cneverStarted\"\\\n\nOomContext\x12#\n\rerror_message\x18\x01 \x01(\tR\x0c\x65rrorMessage\x12)\n\x10\x66\x61il_immediately\x18\x02 \x01(\x08R\x0f\x66\x61ilImmediately\"\xef\x03\n\tJobConfig\x12\x1f\n\x0bjvm_options\x18\x02 \x03(\tR\njvmOptions\x12(\n\x10\x63ode_search_path\x18\x03 \x03(\tR\x0e\x63odeSearchPath\x12\x41\n\x10runtime_env_info\x18\x04 \x01(\x0b\x32\x17.ray.rpc.RuntimeEnvInfoR\x0eruntimeEnvInfo\x12#\n\rray_namespace\x18\x05 \x01(\tR\x0crayNamespace\x12<\n\x08metadata\x18\x06 \x03(\x0b\x32 .ray.rpc.JobConfig.MetadataEntryR\x08metadata\x12V\n\x16\x64\x65\x66\x61ult_actor_lifetime\x18\x07 \x01(\x0e\x32 .ray.rpc.JobConfig.ActorLifetimeR\x14\x64\x65\x66\x61ultActorLifetime\x12+\n\x12py_driver_sys_path\x18\x08 \x03(\tR\x0fpyDriverSysPath\x1a;\n\rMetadataEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"/\n\rActorLifetime\x12\x0c\n\x08\x44\x45TACHED\x10\x00\x12\x10\n\x0cNON_DETACHED\x10\x01\"\x9b\r\n\x08TaskSpec\x12%\n\x04type\x18\x01 \x01(\x0e\x32\x11.ray.rpc.TaskTypeR\x04type\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12-\n\x08language\x18\x03 \x01(\x0e\x32\x11.ray.rpc.LanguageR\x08language\x12L\n\x13\x66unction_descriptor\x18\x04 \x01(\x0b\x32\x1b.ray.rpc.FunctionDescriptorR\x12\x66unctionDescriptor\x12\x15\n\x06job_id\x18\x05 \x01(\x0cR\x05jobId\x12\x17\n\x07task_id\x18\x06 \x01(\x0cR\x06taskId\x12$\n\x0eparent_task_id\x18\x07 \x01(\x0cR\x0cparentTaskId\x12%\n\x0eparent_counter\x18\x08 \x01(\x04R\rparentCounter\x12\x1b\n\tcaller_id\x18\t \x01(\x0cR\x08\x63\x61llerId\x12\x37\n\x0e\x63\x61ller_address\x18\n \x01(\x0b\x32\x10.ray.rpc.AddressR\rcallerAddress\x12$\n\x04\x61rgs\x18\x0b \x03(\x0b\x32\x10.ray.rpc.TaskArgR\x04\x61rgs\x12\x1f\n\x0bnum_returns\x18\x0c \x01(\x04R\nnumReturns\x12W\n\x12required_resources\x18\r \x03(\x0b\x32(.ray.rpc.TaskSpec.RequiredResourcesEntryR\x11requiredResources\x12s\n\x1crequired_placement_resources\x18\x0e \x03(\x0b\x32\x31.ray.rpc.TaskSpec.RequiredPlacementResourcesEntryR\x1arequiredPlacementResources\x12W\n\x18\x61\x63tor_creation_task_spec\x18\x0f \x01(\x0b\x32\x1e.ray.rpc.ActorCreationTaskSpecR\x15\x61\x63torCreationTaskSpec\x12>\n\x0f\x61\x63tor_task_spec\x18\x10 \x01(\x0b\x32\x16.ray.rpc.ActorTaskSpecR\ractorTaskSpec\x12\x1f\n\x0bmax_retries\x18\x11 \x01(\x05R\nmaxRetries\x12%\n\x0eskip_execution\x18\x15 \x01(\x08R\rskipExecution\x12/\n\x13\x64\x65\x62ugger_breakpoint\x18\x16 \x01(\x0cR\x12\x64\x65\x62uggerBreakpoint\x12\x41\n\x10runtime_env_info\x18\x17 \x01(\x0b\x32\x17.ray.rpc.RuntimeEnvInfoR\x0eruntimeEnvInfo\x12\x34\n\x16\x63oncurrency_group_name\x18\x18 \x01(\tR\x14\x63oncurrencyGroupName\x12)\n\x10retry_exceptions\x18\x19 \x01(\x08R\x0fretryExceptions\x12O\n$serialized_retry_exception_allowlist\x18\x1a \x01(\x0cR!serializedRetryExceptionAllowlist\x12\x14\n\x05\x64\x65pth\x18\x1b \x01(\x03R\x05\x64\x65pth\x12L\n\x13scheduling_strategy\x18\x1c \x01(\x0b\x32\x1b.ray.rpc.SchedulingStrategyR\x12schedulingStrategy\x12%\n\x0e\x61ttempt_number\x18\x1d \x01(\x04R\rattemptNumber\x12\'\n\x0freturns_dynamic\x18\x1e \x01(\x08R\x0ereturnsDynamic\x12,\n\x12\x64ynamic_return_ids\x18\x1f \x03(\x0cR\x10\x64ynamicReturnIds\x12\x36\n\njob_config\x18  \x01(\x0b\x32\x12.ray.rpc.JobConfigH\x00R\tjobConfig\x88\x01\x01\x12*\n\x11submitter_task_id\x18! \x01(\x0cR\x0fsubmitterTaskId\x1a\x44\n\x16RequiredResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x1aM\n\x1fRequiredPlacementResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x42\r\n\x0b_job_config\"\xc4\x05\n\rTaskInfoEntry\x12%\n\x04type\x18\x01 \x01(\x0e\x32\x11.ray.rpc.TaskTypeR\x04type\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12-\n\x08language\x18\x03 \x01(\x0e\x32\x11.ray.rpc.LanguageR\x08language\x12+\n\x12\x66unc_or_class_name\x18\x04 \x01(\tR\x0f\x66uncOrClassName\x12>\n\x10scheduling_state\x18\x05 \x01(\x0e\x32\x13.ray.rpc.TaskStatusR\x0fschedulingState\x12\x15\n\x06job_id\x18\x06 \x01(\x0cR\x05jobId\x12\x17\n\x07task_id\x18\x07 \x01(\x0cR\x06taskId\x12$\n\x0eparent_task_id\x18\x08 \x01(\x0cR\x0cparentTaskId\x12\\\n\x12required_resources\x18\r \x03(\x0b\x32-.ray.rpc.TaskInfoEntry.RequiredResourcesEntryR\x11requiredResources\x12\x41\n\x10runtime_env_info\x18\x17 \x01(\x0b\x32\x17.ray.rpc.RuntimeEnvInfoR\x0eruntimeEnvInfo\x12\x1c\n\x07node_id\x18\x18 \x01(\x0cH\x00R\x06nodeId\x88\x01\x01\x12\x1e\n\x08\x61\x63tor_id\x18\x19 \x01(\x0cH\x01R\x07\x61\x63torId\x88\x01\x01\x12\x31\n\x12placement_group_id\x18\x1a \x01(\x0cH\x02R\x10placementGroupId\x88\x01\x01\x1a\x44\n\x16RequiredResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\x42\n\n\x08_node_idB\x0b\n\t_actor_idB\x15\n\x13_placement_group_id\"\xd2\x02\n\x06\x42undle\x12=\n\tbundle_id\x18\x01 \x01(\x0b\x32 .ray.rpc.Bundle.BundleIdentifierR\x08\x62undleId\x12I\n\x0eunit_resources\x18\x02 \x03(\x0b\x32\".ray.rpc.Bundle.UnitResourcesEntryR\runitResources\x12\x17\n\x07node_id\x18\x03 \x01(\x0cR\x06nodeId\x1a\x63\n\x10\x42undleIdentifier\x12,\n\x12placement_group_id\x18\x01 \x01(\x0cR\x10placementGroupId\x12!\n\x0c\x62undle_index\x18\x02 \x01(\x05R\x0b\x62undleIndex\x1a@\n\x12UnitResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\x01R\x05value:\x02\x38\x01\"\xbc\x03\n\x12PlacementGroupSpec\x12,\n\x12placement_group_id\x18\x01 \x01(\x0cR\x10placementGroupId\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12)\n\x07\x62undles\x18\x03 \x03(\x0b\x32\x0f.ray.rpc.BundleR\x07\x62undles\x12\x36\n\x08strategy\x18\x04 \x01(\x0e\x32\x1a.ray.rpc.PlacementStrategyR\x08strategy\x12$\n\x0e\x63reator_job_id\x18\x05 \x01(\x0cR\x0c\x63reatorJobId\x12(\n\x10\x63reator_actor_id\x18\x06 \x01(\x0cR\x0e\x63reatorActorId\x12(\n\x10\x63reator_job_dead\x18\x07 \x01(\x08R\x0e\x63reatorJobDead\x12,\n\x12\x63reator_actor_dead\x18\x08 \x01(\x08R\x10\x63reatorActorDead\x12\x1f\n\x0bis_detached\x18\t \x01(\x08R\nisDetached\x12\x38\n\x19max_cpu_fraction_per_node\x18\n \x01(\x01R\x15maxCpuFractionPerNode\"\x82\x01\n\x0fObjectReference\x12\x1b\n\tobject_id\x18\x01 \x01(\x0cR\x08objectId\x12\x35\n\rowner_address\x18\x02 \x01(\x0b\x32\x10.ray.rpc.AddressR\x0cownerAddress\x12\x1b\n\tcall_site\x18\x03 \x01(\tR\x08\x63\x61llSite\"\xbf\x02\n\x14ObjectReferenceCount\x12\x36\n\treference\x18\x01 \x01(\x0b\x32\x18.ray.rpc.ObjectReferenceR\treference\x12\"\n\rhas_local_ref\x18\x02 \x01(\x08R\x0bhasLocalRef\x12.\n\tborrowers\x18\x03 \x03(\x0b\x32\x10.ray.rpc.AddressR\tborrowers\x12\x44\n\x11stored_in_objects\x18\x04 \x03(\x0b\x32\x18.ray.rpc.ObjectReferenceR\x0fstoredInObjects\x12\x39\n\x19\x63ontained_in_borrowed_ids\x18\x05 \x03(\x0cR\x16\x63ontainedInBorrowedIds\x12\x1a\n\x08\x63ontains\x18\x06 \x03(\x0cR\x08\x63ontains\"\xbc\x01\n\x07TaskArg\x12\x37\n\nobject_ref\x18\x01 \x01(\x0b\x32\x18.ray.rpc.ObjectReferenceR\tobjectRef\x12\x12\n\x04\x64\x61ta\x18\x02 \x01(\x0cR\x04\x64\x61ta\x12\x1a\n\x08metadata\x18\x03 \x01(\x0cR\x08metadata\x12H\n\x13nested_inlined_refs\x18\x04 \x03(\x0b\x32\x18.ray.rpc.ObjectReferenceR\x11nestedInlinedRefs\"\xe8\x04\n\x15\x41\x63torCreationTaskSpec\x12\x19\n\x08\x61\x63tor_id\x18\x02 \x01(\x0cR\x07\x61\x63torId\x12,\n\x12max_actor_restarts\x18\x03 \x01(\x03R\x10maxActorRestarts\x12(\n\x10max_task_retries\x18\x04 \x01(\x03R\x0emaxTaskRetries\x12\x34\n\x16\x64ynamic_worker_options\x18\x05 \x03(\tR\x14\x64ynamicWorkerOptions\x12\'\n\x0fmax_concurrency\x18\x06 \x01(\x05R\x0emaxConcurrency\x12\x1f\n\x0bis_detached\x18\x07 \x01(\x08R\nisDetached\x12\x12\n\x04name\x18\x08 \x01(\tR\x04name\x12#\n\rray_namespace\x18\t \x01(\tR\x0crayNamespace\x12\x1d\n\nis_asyncio\x18\n \x01(\x08R\tisAsyncio\x12%\n\x0e\x65xtension_data\x18\x0b \x01(\tR\rextensionData\x12\x36\n\x17serialized_actor_handle\x18\x0c \x01(\x0cR\x15serializedActorHandle\x12H\n\x12\x63oncurrency_groups\x18\r \x03(\x0b\x32\x19.ray.rpc.ConcurrencyGroupR\x11\x63oncurrencyGroups\x12/\n\x14\x65xecute_out_of_order\x18\x0e \x01(\x08R\x11\x65xecuteOutOfOrder\x12*\n\x11max_pending_calls\x18\x0f \x01(\x05R\x0fmaxPendingCalls\"\x93\x01\n\rActorTaskSpec\x12\x19\n\x08\x61\x63tor_id\x18\x02 \x01(\x0cR\x07\x61\x63torId\x12\x42\n\x1e\x61\x63tor_creation_dummy_object_id\x18\x04 \x01(\x0cR\x1a\x61\x63torCreationDummyObjectId\x12#\n\ractor_counter\x18\x05 \x01(\x04R\x0c\x61\x63torCounter\"6\n\x04Task\x12.\n\ttask_spec\x18\x01 \x01(\x0b\x32\x11.ray.rpc.TaskSpecR\x08taskSpec\">\n\nResourceId\x12\x14\n\x05index\x18\x01 \x01(\x03R\x05index\x12\x1a\n\x08quantity\x18\x02 \x01(\x01R\x08quantity\"^\n\x10ResourceMapEntry\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x36\n\x0cresource_ids\x18\x02 \x03(\x0b\x32\x13.ray.rpc.ResourceIdR\x0bresourceIds\"\xf4\x03\n\x08ViewData\x12\x1b\n\tview_name\x18\x01 \x01(\tR\x08viewName\x12\x35\n\x08measures\x18\x02 \x03(\x0b\x32\x19.ray.rpc.ViewData.MeasureR\x08measures\x1a\x93\x03\n\x07Measure\x12\x12\n\x04tags\x18\x01 \x01(\tR\x04tags\x12\x1b\n\tint_value\x18\x02 \x01(\x03R\x08intValue\x12!\n\x0c\x64ouble_value\x18\x03 \x01(\x01R\x0b\x64oubleValue\x12)\n\x10\x64istribution_min\x18\x04 \x01(\x01R\x0f\x64istributionMin\x12+\n\x11\x64istribution_mean\x18\x05 \x01(\x01R\x10\x64istributionMean\x12)\n\x10\x64istribution_max\x18\x06 \x01(\x01R\x0f\x64istributionMax\x12-\n\x12\x64istribution_count\x18\x07 \x01(\x01R\x11\x64istributionCount\x12\x44\n\x1e\x64istribution_bucket_boundaries\x18\x08 \x03(\x01R\x1c\x64istributionBucketBoundaries\x12<\n\x1a\x64istribution_bucket_counts\x18\t \x03(\x01R\x18\x64istributionBucketCounts\"\x80\x03\n\rObjectRefInfo\x12\x1b\n\tobject_id\x18\x01 \x01(\x0cR\x08objectId\x12\x1b\n\tcall_site\x18\x02 \x01(\tR\x08\x63\x61llSite\x12\x1f\n\x0bobject_size\x18\x03 \x01(\x03R\nobjectSize\x12&\n\x0flocal_ref_count\x18\x04 \x01(\x03R\rlocalRefCount\x12\x37\n\x18submitted_task_ref_count\x18\x05 \x01(\x03R\x15submittedTaskRefCount\x12,\n\x12\x63ontained_in_owned\x18\x06 \x03(\x0cR\x10\x63ontainedInOwned\x12(\n\x10pinned_in_memory\x18\x07 \x01(\x08R\x0epinnedInMemory\x12\x34\n\x0btask_status\x18\x08 \x01(\x0e\x32\x13.ray.rpc.TaskStatusR\ntaskStatus\x12%\n\x0e\x61ttempt_number\x18\t \x01(\x04R\rattemptNumber\"\xab\x01\n\x13ResourceAllocations\x12P\n\x0eresource_slots\x18\x01 \x03(\x0b\x32).ray.rpc.ResourceAllocations.ResourceSlotR\rresourceSlots\x1a\x42\n\x0cResourceSlot\x12\x12\n\x04slot\x18\x01 \x01(\x03R\x04slot\x12\x1e\n\nallocation\x18\x02 \x01(\x01R\nallocation\"\xc2\x08\n\x0f\x43oreWorkerStats\x12*\n\x11\x63urrent_task_desc\x18\x01 \x01(\tR\x0f\x63urrentTaskDesc\x12*\n\x11num_pending_tasks\x18\x02 \x01(\x05R\x0fnumPendingTasks\x12\x36\n\x18num_object_refs_in_scope\x18\x03 \x01(\x05R\x14numObjectRefsInScope\x12\x1d\n\nip_address\x18\x07 \x01(\tR\tipAddress\x12\x12\n\x04port\x18\x08 \x01(\x03R\x04port\x12\x19\n\x08\x61\x63tor_id\x18\t \x01(\x0cR\x07\x61\x63torId\x12R\n\x0eused_resources\x18\n \x03(\x0b\x32+.ray.rpc.CoreWorkerStats.UsedResourcesEntryR\rusedResources\x12O\n\rwebui_display\x18\x0b \x03(\x0b\x32*.ray.rpc.CoreWorkerStats.WebuiDisplayEntryR\x0cwebuiDisplay\x12\"\n\rnum_in_plasma\x18\x0c \x01(\x05R\x0bnumInPlasma\x12*\n\x11num_local_objects\x18\r \x01(\x05R\x0fnumLocalObjects\x12\x37\n\x18used_object_store_memory\x18\x0e \x01(\x03R\x15usedObjectStoreMemory\x12*\n\x11task_queue_length\x18\x0f \x01(\x05R\x0ftaskQueueLength\x12,\n\x12num_executed_tasks\x18\x10 \x01(\x05R\x10numExecutedTasks\x12\x1f\n\x0b\x61\x63tor_title\x18\x11 \x01(\tR\nactorTitle\x12\x37\n\x0bobject_refs\x18\x12 \x03(\x0b\x32\x16.ray.rpc.ObjectRefInfoR\nobjectRefs\x12\x15\n\x06job_id\x18\x13 \x01(\x0cR\x05jobId\x12\x1b\n\tworker_id\x18\x14 \x01(\x0cR\x08workerId\x12-\n\x08language\x18\x15 \x01(\x0e\x32\x11.ray.rpc.LanguageR\x08language\x12\x10\n\x03pid\x18\x16 \x01(\rR\x03pid\x12\x34\n\x0bworker_type\x18\x17 \x01(\x0e\x32\x13.ray.rpc.WorkerTypeR\nworkerType\x12#\n\robjects_total\x18\x18 \x01(\x03R\x0cobjectsTotal\x1a^\n\x12UsedResourcesEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x32\n\x05value\x18\x02 \x01(\x0b\x32\x1c.ray.rpc.ResourceAllocationsR\x05value:\x02\x38\x01\x1a?\n\x11WebuiDisplayEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"\'\n\x11NodeResourceUsage\x12\x12\n\x04json\x18\x01 \x01(\tR\x04json\"\x87\x02\n\x0bMetricPoint\x12\x1f\n\x0bmetric_name\x18\x01 \x01(\tR\nmetricName\x12\x1c\n\ttimestamp\x18\x02 \x01(\x03R\ttimestamp\x12\x14\n\x05value\x18\x03 \x01(\x01R\x05value\x12\x32\n\x04tags\x18\x04 \x03(\x0b\x32\x1e.ray.rpc.MetricPoint.TagsEntryR\x04tags\x12 \n\x0b\x64\x65scription\x18\x05 \x01(\tR\x0b\x64\x65scription\x12\x14\n\x05units\x18\x06 \x01(\tR\x05units\x1a\x37\n\tTagsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\"I\n\x0eNamedActorInfo\x12#\n\rray_namespace\x18\x01 \x01(\tR\x0crayNamespace\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name*)\n\x08Language\x12\n\n\x06PYTHON\x10\x00\x12\x08\n\x04JAVA\x10\x01\x12\x07\n\x03\x43PP\x10\x02*J\n\nWorkerType\x12\n\n\x06WORKER\x10\x00\x12\n\n\x06\x44RIVER\x10\x01\x12\x10\n\x0cSPILL_WORKER\x10\x02\x12\x12\n\x0eRESTORE_WORKER\x10\x03*U\n\x08TaskType\x12\x0f\n\x0bNORMAL_TASK\x10\x00\x12\x17\n\x13\x41\x43TOR_CREATION_TASK\x10\x01\x12\x0e\n\nACTOR_TASK\x10\x02\x12\x0f\n\x0b\x44RIVER_TASK\x10\x03*\x84\x05\n\tErrorType\x12\x0f\n\x0bWORKER_DIED\x10\x00\x12\x0e\n\nACTOR_DIED\x10\x01\x12\x1c\n\x18OBJECT_UNRECONSTRUCTABLE\x10\x02\x12\x1c\n\x18TASK_EXECUTION_EXCEPTION\x10\x03\x12\x14\n\x10OBJECT_IN_PLASMA\x10\x04\x12\x12\n\x0eTASK_CANCELLED\x10\x05\x12\x19\n\x15\x41\x43TOR_CREATION_FAILED\x10\x06\x12\x1c\n\x18RUNTIME_ENV_SETUP_FAILED\x10\x07\x12\x0f\n\x0bOBJECT_LOST\x10\x08\x12\x0e\n\nOWNER_DIED\x10\t\x12\x12\n\x0eOBJECT_DELETED\x10\n\x12 \n\x1c\x44\x45PENDENCY_RESOLUTION_FAILED\x10\x0b\x12\x32\n.OBJECT_UNRECONSTRUCTABLE_MAX_ATTEMPTS_EXCEEDED\x10\x0c\x12,\n(OBJECT_UNRECONSTRUCTABLE_LINEAGE_EVICTED\x10\r\x12\x1a\n\x16OBJECT_FETCH_TIMED_OUT\x10\x0e\x12\x15\n\x11LOCAL_RAYLET_DIED\x10\x0f\x12 \n\x1cTASK_PLACEMENT_GROUP_REMOVED\x10\x10\x12!\n\x1d\x41\x43TOR_PLACEMENT_GROUP_REMOVED\x10\x11\x12\x1c\n\x18TASK_UNSCHEDULABLE_ERROR\x10\x12\x12\x1d\n\x19\x41\x43TOR_UNSCHEDULABLE_ERROR\x10\x13\x12\x15\n\x11OUT_OF_DISK_ERROR\x10\x14\x12\x10\n\x0cOBJECT_FREED\x10\x15\x12\x11\n\rOUT_OF_MEMORY\x10\x16\x12\r\n\tNODE_DIED\x10\x17*\xf4\x01\n\nTaskStatus\x12\x07\n\x03NIL\x10\x00\x12\x16\n\x12PENDING_ARGS_AVAIL\x10\x01\x12\x1b\n\x17PENDING_NODE_ASSIGNMENT\x10\x02\x12\x1f\n\x1bPENDING_OBJ_STORE_MEM_AVAIL\x10\x03\x12\x16\n\x12PENDING_ARGS_FETCH\x10\x04\x12\x17\n\x13SUBMITTED_TO_WORKER\x10\x05\x12\x0b\n\x07RUNNING\x10\x06\x12\x16\n\x12RUNNING_IN_RAY_GET\x10\x07\x12\x17\n\x13RUNNING_IN_RAY_WAIT\x10\x08\x12\x0c\n\x08\x46INISHED\x10\t\x12\n\n\x06\x46\x41ILED\x10\n*|\n\x0eWorkerExitType\x12\x10\n\x0cSYSTEM_ERROR\x10\x00\x12\x18\n\x14INTENDED_SYSTEM_EXIT\x10\x01\x12\x0e\n\nUSER_ERROR\x10\x02\x12\x16\n\x12INTENDED_USER_EXIT\x10\x03\x12\x16\n\x12NODE_OUT_OF_MEMORY\x10\x04*M\n\x11PlacementStrategy\x12\x08\n\x04PACK\x10\x00\x12\n\n\x06SPREAD\x10\x01\x12\x0f\n\x0bSTRICT_PACK\x10\x02\x12\x11\n\rSTRICT_SPREAD\x10\x03\x42\x1d\n\x18io.ray.runtime.generated\xf8\x01\x01\x62\x06proto3')

_LANGUAGE = DESCRIPTOR.enum_types_by_name['Language']
Language = enum_type_wrapper.EnumTypeWrapper(_LANGUAGE)
_WORKERTYPE = DESCRIPTOR.enum_types_by_name['WorkerType']
WorkerType = enum_type_wrapper.EnumTypeWrapper(_WORKERTYPE)
_TASKTYPE = DESCRIPTOR.enum_types_by_name['TaskType']
TaskType = enum_type_wrapper.EnumTypeWrapper(_TASKTYPE)
_ERRORTYPE = DESCRIPTOR.enum_types_by_name['ErrorType']
ErrorType = enum_type_wrapper.EnumTypeWrapper(_ERRORTYPE)
_TASKSTATUS = DESCRIPTOR.enum_types_by_name['TaskStatus']
TaskStatus = enum_type_wrapper.EnumTypeWrapper(_TASKSTATUS)
_WORKEREXITTYPE = DESCRIPTOR.enum_types_by_name['WorkerExitType']
WorkerExitType = enum_type_wrapper.EnumTypeWrapper(_WORKEREXITTYPE)
_PLACEMENTSTRATEGY = DESCRIPTOR.enum_types_by_name['PlacementStrategy']
PlacementStrategy = enum_type_wrapper.EnumTypeWrapper(_PLACEMENTSTRATEGY)
PYTHON = 0
JAVA = 1
CPP = 2
WORKER = 0
DRIVER = 1
SPILL_WORKER = 2
RESTORE_WORKER = 3
NORMAL_TASK = 0
ACTOR_CREATION_TASK = 1
ACTOR_TASK = 2
DRIVER_TASK = 3
WORKER_DIED = 0
ACTOR_DIED = 1
OBJECT_UNRECONSTRUCTABLE = 2
TASK_EXECUTION_EXCEPTION = 3
OBJECT_IN_PLASMA = 4
TASK_CANCELLED = 5
ACTOR_CREATION_FAILED = 6
RUNTIME_ENV_SETUP_FAILED = 7
OBJECT_LOST = 8
OWNER_DIED = 9
OBJECT_DELETED = 10
DEPENDENCY_RESOLUTION_FAILED = 11
OBJECT_UNRECONSTRUCTABLE_MAX_ATTEMPTS_EXCEEDED = 12
OBJECT_UNRECONSTRUCTABLE_LINEAGE_EVICTED = 13
OBJECT_FETCH_TIMED_OUT = 14
LOCAL_RAYLET_DIED = 15
TASK_PLACEMENT_GROUP_REMOVED = 16
ACTOR_PLACEMENT_GROUP_REMOVED = 17
TASK_UNSCHEDULABLE_ERROR = 18
ACTOR_UNSCHEDULABLE_ERROR = 19
OUT_OF_DISK_ERROR = 20
OBJECT_FREED = 21
OUT_OF_MEMORY = 22
NODE_DIED = 23
NIL = 0
PENDING_ARGS_AVAIL = 1
PENDING_NODE_ASSIGNMENT = 2
PENDING_OBJ_STORE_MEM_AVAIL = 3
PENDING_ARGS_FETCH = 4
SUBMITTED_TO_WORKER = 5
RUNNING = 6
RUNNING_IN_RAY_GET = 7
RUNNING_IN_RAY_WAIT = 8
FINISHED = 9
FAILED = 10
SYSTEM_ERROR = 0
INTENDED_SYSTEM_EXIT = 1
USER_ERROR = 2
INTENDED_USER_EXIT = 3
NODE_OUT_OF_MEMORY = 4
PACK = 0
SPREAD = 1
STRICT_PACK = 2
STRICT_SPREAD = 3


_NODEAFFINITYSCHEDULINGSTRATEGY = DESCRIPTOR.message_types_by_name['NodeAffinitySchedulingStrategy']
_PLACEMENTGROUPSCHEDULINGSTRATEGY = DESCRIPTOR.message_types_by_name['PlacementGroupSchedulingStrategy']
_DEFAULTSCHEDULINGSTRATEGY = DESCRIPTOR.message_types_by_name['DefaultSchedulingStrategy']
_SPREADSCHEDULINGSTRATEGY = DESCRIPTOR.message_types_by_name['SpreadSchedulingStrategy']
_SCHEDULINGSTRATEGY = DESCRIPTOR.message_types_by_name['SchedulingStrategy']
_ADDRESS = DESCRIPTOR.message_types_by_name['Address']
_JAVAFUNCTIONDESCRIPTOR = DESCRIPTOR.message_types_by_name['JavaFunctionDescriptor']
_PYTHONFUNCTIONDESCRIPTOR = DESCRIPTOR.message_types_by_name['PythonFunctionDescriptor']
_CPPFUNCTIONDESCRIPTOR = DESCRIPTOR.message_types_by_name['CppFunctionDescriptor']
_FUNCTIONDESCRIPTOR = DESCRIPTOR.message_types_by_name['FunctionDescriptor']
_CONCURRENCYGROUP = DESCRIPTOR.message_types_by_name['ConcurrencyGroup']
_RAYERRORINFO = DESCRIPTOR.message_types_by_name['RayErrorInfo']
_OUTOFMEMORYERRORCONTEXT = DESCRIPTOR.message_types_by_name['OutOfMemoryErrorContext']
_NODEDIEDERRORCONTEXT = DESCRIPTOR.message_types_by_name['NodeDiedErrorContext']
_RAYEXCEPTION = DESCRIPTOR.message_types_by_name['RayException']
_ACTORDEATHCAUSE = DESCRIPTOR.message_types_by_name['ActorDeathCause']
_RUNTIMEENVFAILEDCONTEXT = DESCRIPTOR.message_types_by_name['RuntimeEnvFailedContext']
_ACTORUNSCHEDULABLECONTEXT = DESCRIPTOR.message_types_by_name['ActorUnschedulableContext']
_ACTORDIEDERRORCONTEXT = DESCRIPTOR.message_types_by_name['ActorDiedErrorContext']
_OOMCONTEXT = DESCRIPTOR.message_types_by_name['OomContext']
_JOBCONFIG = DESCRIPTOR.message_types_by_name['JobConfig']
_JOBCONFIG_METADATAENTRY = _JOBCONFIG.nested_types_by_name['MetadataEntry']
_TASKSPEC = DESCRIPTOR.message_types_by_name['TaskSpec']
_TASKSPEC_REQUIREDRESOURCESENTRY = _TASKSPEC.nested_types_by_name['RequiredResourcesEntry']
_TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY = _TASKSPEC.nested_types_by_name['RequiredPlacementResourcesEntry']
_TASKINFOENTRY = DESCRIPTOR.message_types_by_name['TaskInfoEntry']
_TASKINFOENTRY_REQUIREDRESOURCESENTRY = _TASKINFOENTRY.nested_types_by_name['RequiredResourcesEntry']
_BUNDLE = DESCRIPTOR.message_types_by_name['Bundle']
_BUNDLE_BUNDLEIDENTIFIER = _BUNDLE.nested_types_by_name['BundleIdentifier']
_BUNDLE_UNITRESOURCESENTRY = _BUNDLE.nested_types_by_name['UnitResourcesEntry']
_PLACEMENTGROUPSPEC = DESCRIPTOR.message_types_by_name['PlacementGroupSpec']
_OBJECTREFERENCE = DESCRIPTOR.message_types_by_name['ObjectReference']
_OBJECTREFERENCECOUNT = DESCRIPTOR.message_types_by_name['ObjectReferenceCount']
_TASKARG = DESCRIPTOR.message_types_by_name['TaskArg']
_ACTORCREATIONTASKSPEC = DESCRIPTOR.message_types_by_name['ActorCreationTaskSpec']
_ACTORTASKSPEC = DESCRIPTOR.message_types_by_name['ActorTaskSpec']
_TASK = DESCRIPTOR.message_types_by_name['Task']
_RESOURCEID = DESCRIPTOR.message_types_by_name['ResourceId']
_RESOURCEMAPENTRY = DESCRIPTOR.message_types_by_name['ResourceMapEntry']
_VIEWDATA = DESCRIPTOR.message_types_by_name['ViewData']
_VIEWDATA_MEASURE = _VIEWDATA.nested_types_by_name['Measure']
_OBJECTREFINFO = DESCRIPTOR.message_types_by_name['ObjectRefInfo']
_RESOURCEALLOCATIONS = DESCRIPTOR.message_types_by_name['ResourceAllocations']
_RESOURCEALLOCATIONS_RESOURCESLOT = _RESOURCEALLOCATIONS.nested_types_by_name['ResourceSlot']
_COREWORKERSTATS = DESCRIPTOR.message_types_by_name['CoreWorkerStats']
_COREWORKERSTATS_USEDRESOURCESENTRY = _COREWORKERSTATS.nested_types_by_name['UsedResourcesEntry']
_COREWORKERSTATS_WEBUIDISPLAYENTRY = _COREWORKERSTATS.nested_types_by_name['WebuiDisplayEntry']
_NODERESOURCEUSAGE = DESCRIPTOR.message_types_by_name['NodeResourceUsage']
_METRICPOINT = DESCRIPTOR.message_types_by_name['MetricPoint']
_METRICPOINT_TAGSENTRY = _METRICPOINT.nested_types_by_name['TagsEntry']
_NAMEDACTORINFO = DESCRIPTOR.message_types_by_name['NamedActorInfo']
_JOBCONFIG_ACTORLIFETIME = _JOBCONFIG.enum_types_by_name['ActorLifetime']
NodeAffinitySchedulingStrategy = _reflection.GeneratedProtocolMessageType('NodeAffinitySchedulingStrategy', (_message.Message,), {
  'DESCRIPTOR' : _NODEAFFINITYSCHEDULINGSTRATEGY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NodeAffinitySchedulingStrategy)
  })
_sym_db.RegisterMessage(NodeAffinitySchedulingStrategy)

PlacementGroupSchedulingStrategy = _reflection.GeneratedProtocolMessageType('PlacementGroupSchedulingStrategy', (_message.Message,), {
  'DESCRIPTOR' : _PLACEMENTGROUPSCHEDULINGSTRATEGY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PlacementGroupSchedulingStrategy)
  })
_sym_db.RegisterMessage(PlacementGroupSchedulingStrategy)

DefaultSchedulingStrategy = _reflection.GeneratedProtocolMessageType('DefaultSchedulingStrategy', (_message.Message,), {
  'DESCRIPTOR' : _DEFAULTSCHEDULINGSTRATEGY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.DefaultSchedulingStrategy)
  })
_sym_db.RegisterMessage(DefaultSchedulingStrategy)

SpreadSchedulingStrategy = _reflection.GeneratedProtocolMessageType('SpreadSchedulingStrategy', (_message.Message,), {
  'DESCRIPTOR' : _SPREADSCHEDULINGSTRATEGY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.SpreadSchedulingStrategy)
  })
_sym_db.RegisterMessage(SpreadSchedulingStrategy)

SchedulingStrategy = _reflection.GeneratedProtocolMessageType('SchedulingStrategy', (_message.Message,), {
  'DESCRIPTOR' : _SCHEDULINGSTRATEGY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.SchedulingStrategy)
  })
_sym_db.RegisterMessage(SchedulingStrategy)

Address = _reflection.GeneratedProtocolMessageType('Address', (_message.Message,), {
  'DESCRIPTOR' : _ADDRESS,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.Address)
  })
_sym_db.RegisterMessage(Address)

JavaFunctionDescriptor = _reflection.GeneratedProtocolMessageType('JavaFunctionDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _JAVAFUNCTIONDESCRIPTOR,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.JavaFunctionDescriptor)
  })
_sym_db.RegisterMessage(JavaFunctionDescriptor)

PythonFunctionDescriptor = _reflection.GeneratedProtocolMessageType('PythonFunctionDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _PYTHONFUNCTIONDESCRIPTOR,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PythonFunctionDescriptor)
  })
_sym_db.RegisterMessage(PythonFunctionDescriptor)

CppFunctionDescriptor = _reflection.GeneratedProtocolMessageType('CppFunctionDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _CPPFUNCTIONDESCRIPTOR,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CppFunctionDescriptor)
  })
_sym_db.RegisterMessage(CppFunctionDescriptor)

FunctionDescriptor = _reflection.GeneratedProtocolMessageType('FunctionDescriptor', (_message.Message,), {
  'DESCRIPTOR' : _FUNCTIONDESCRIPTOR,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.FunctionDescriptor)
  })
_sym_db.RegisterMessage(FunctionDescriptor)

ConcurrencyGroup = _reflection.GeneratedProtocolMessageType('ConcurrencyGroup', (_message.Message,), {
  'DESCRIPTOR' : _CONCURRENCYGROUP,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ConcurrencyGroup)
  })
_sym_db.RegisterMessage(ConcurrencyGroup)

RayErrorInfo = _reflection.GeneratedProtocolMessageType('RayErrorInfo', (_message.Message,), {
  'DESCRIPTOR' : _RAYERRORINFO,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RayErrorInfo)
  })
_sym_db.RegisterMessage(RayErrorInfo)

OutOfMemoryErrorContext = _reflection.GeneratedProtocolMessageType('OutOfMemoryErrorContext', (_message.Message,), {
  'DESCRIPTOR' : _OUTOFMEMORYERRORCONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.OutOfMemoryErrorContext)
  })
_sym_db.RegisterMessage(OutOfMemoryErrorContext)

NodeDiedErrorContext = _reflection.GeneratedProtocolMessageType('NodeDiedErrorContext', (_message.Message,), {
  'DESCRIPTOR' : _NODEDIEDERRORCONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NodeDiedErrorContext)
  })
_sym_db.RegisterMessage(NodeDiedErrorContext)

RayException = _reflection.GeneratedProtocolMessageType('RayException', (_message.Message,), {
  'DESCRIPTOR' : _RAYEXCEPTION,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RayException)
  })
_sym_db.RegisterMessage(RayException)

ActorDeathCause = _reflection.GeneratedProtocolMessageType('ActorDeathCause', (_message.Message,), {
  'DESCRIPTOR' : _ACTORDEATHCAUSE,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorDeathCause)
  })
_sym_db.RegisterMessage(ActorDeathCause)

RuntimeEnvFailedContext = _reflection.GeneratedProtocolMessageType('RuntimeEnvFailedContext', (_message.Message,), {
  'DESCRIPTOR' : _RUNTIMEENVFAILEDCONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.RuntimeEnvFailedContext)
  })
_sym_db.RegisterMessage(RuntimeEnvFailedContext)

ActorUnschedulableContext = _reflection.GeneratedProtocolMessageType('ActorUnschedulableContext', (_message.Message,), {
  'DESCRIPTOR' : _ACTORUNSCHEDULABLECONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorUnschedulableContext)
  })
_sym_db.RegisterMessage(ActorUnschedulableContext)

ActorDiedErrorContext = _reflection.GeneratedProtocolMessageType('ActorDiedErrorContext', (_message.Message,), {
  'DESCRIPTOR' : _ACTORDIEDERRORCONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorDiedErrorContext)
  })
_sym_db.RegisterMessage(ActorDiedErrorContext)

OomContext = _reflection.GeneratedProtocolMessageType('OomContext', (_message.Message,), {
  'DESCRIPTOR' : _OOMCONTEXT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.OomContext)
  })
_sym_db.RegisterMessage(OomContext)

JobConfig = _reflection.GeneratedProtocolMessageType('JobConfig', (_message.Message,), {

  'MetadataEntry' : _reflection.GeneratedProtocolMessageType('MetadataEntry', (_message.Message,), {
    'DESCRIPTOR' : _JOBCONFIG_METADATAENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.JobConfig.MetadataEntry)
    })
  ,
  'DESCRIPTOR' : _JOBCONFIG,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.JobConfig)
  })
_sym_db.RegisterMessage(JobConfig)
_sym_db.RegisterMessage(JobConfig.MetadataEntry)

TaskSpec = _reflection.GeneratedProtocolMessageType('TaskSpec', (_message.Message,), {

  'RequiredResourcesEntry' : _reflection.GeneratedProtocolMessageType('RequiredResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _TASKSPEC_REQUIREDRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.TaskSpec.RequiredResourcesEntry)
    })
  ,

  'RequiredPlacementResourcesEntry' : _reflection.GeneratedProtocolMessageType('RequiredPlacementResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.TaskSpec.RequiredPlacementResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _TASKSPEC,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskSpec)
  })
_sym_db.RegisterMessage(TaskSpec)
_sym_db.RegisterMessage(TaskSpec.RequiredResourcesEntry)
_sym_db.RegisterMessage(TaskSpec.RequiredPlacementResourcesEntry)

TaskInfoEntry = _reflection.GeneratedProtocolMessageType('TaskInfoEntry', (_message.Message,), {

  'RequiredResourcesEntry' : _reflection.GeneratedProtocolMessageType('RequiredResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _TASKINFOENTRY_REQUIREDRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.TaskInfoEntry.RequiredResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _TASKINFOENTRY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskInfoEntry)
  })
_sym_db.RegisterMessage(TaskInfoEntry)
_sym_db.RegisterMessage(TaskInfoEntry.RequiredResourcesEntry)

Bundle = _reflection.GeneratedProtocolMessageType('Bundle', (_message.Message,), {

  'BundleIdentifier' : _reflection.GeneratedProtocolMessageType('BundleIdentifier', (_message.Message,), {
    'DESCRIPTOR' : _BUNDLE_BUNDLEIDENTIFIER,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.Bundle.BundleIdentifier)
    })
  ,

  'UnitResourcesEntry' : _reflection.GeneratedProtocolMessageType('UnitResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _BUNDLE_UNITRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.Bundle.UnitResourcesEntry)
    })
  ,
  'DESCRIPTOR' : _BUNDLE,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.Bundle)
  })
_sym_db.RegisterMessage(Bundle)
_sym_db.RegisterMessage(Bundle.BundleIdentifier)
_sym_db.RegisterMessage(Bundle.UnitResourcesEntry)

PlacementGroupSpec = _reflection.GeneratedProtocolMessageType('PlacementGroupSpec', (_message.Message,), {
  'DESCRIPTOR' : _PLACEMENTGROUPSPEC,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.PlacementGroupSpec)
  })
_sym_db.RegisterMessage(PlacementGroupSpec)

ObjectReference = _reflection.GeneratedProtocolMessageType('ObjectReference', (_message.Message,), {
  'DESCRIPTOR' : _OBJECTREFERENCE,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ObjectReference)
  })
_sym_db.RegisterMessage(ObjectReference)

ObjectReferenceCount = _reflection.GeneratedProtocolMessageType('ObjectReferenceCount', (_message.Message,), {
  'DESCRIPTOR' : _OBJECTREFERENCECOUNT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ObjectReferenceCount)
  })
_sym_db.RegisterMessage(ObjectReferenceCount)

TaskArg = _reflection.GeneratedProtocolMessageType('TaskArg', (_message.Message,), {
  'DESCRIPTOR' : _TASKARG,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.TaskArg)
  })
_sym_db.RegisterMessage(TaskArg)

ActorCreationTaskSpec = _reflection.GeneratedProtocolMessageType('ActorCreationTaskSpec', (_message.Message,), {
  'DESCRIPTOR' : _ACTORCREATIONTASKSPEC,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorCreationTaskSpec)
  })
_sym_db.RegisterMessage(ActorCreationTaskSpec)

ActorTaskSpec = _reflection.GeneratedProtocolMessageType('ActorTaskSpec', (_message.Message,), {
  'DESCRIPTOR' : _ACTORTASKSPEC,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ActorTaskSpec)
  })
_sym_db.RegisterMessage(ActorTaskSpec)

Task = _reflection.GeneratedProtocolMessageType('Task', (_message.Message,), {
  'DESCRIPTOR' : _TASK,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.Task)
  })
_sym_db.RegisterMessage(Task)

ResourceId = _reflection.GeneratedProtocolMessageType('ResourceId', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEID,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceId)
  })
_sym_db.RegisterMessage(ResourceId)

ResourceMapEntry = _reflection.GeneratedProtocolMessageType('ResourceMapEntry', (_message.Message,), {
  'DESCRIPTOR' : _RESOURCEMAPENTRY,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceMapEntry)
  })
_sym_db.RegisterMessage(ResourceMapEntry)

ViewData = _reflection.GeneratedProtocolMessageType('ViewData', (_message.Message,), {

  'Measure' : _reflection.GeneratedProtocolMessageType('Measure', (_message.Message,), {
    'DESCRIPTOR' : _VIEWDATA_MEASURE,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ViewData.Measure)
    })
  ,
  'DESCRIPTOR' : _VIEWDATA,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ViewData)
  })
_sym_db.RegisterMessage(ViewData)
_sym_db.RegisterMessage(ViewData.Measure)

ObjectRefInfo = _reflection.GeneratedProtocolMessageType('ObjectRefInfo', (_message.Message,), {
  'DESCRIPTOR' : _OBJECTREFINFO,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ObjectRefInfo)
  })
_sym_db.RegisterMessage(ObjectRefInfo)

ResourceAllocations = _reflection.GeneratedProtocolMessageType('ResourceAllocations', (_message.Message,), {

  'ResourceSlot' : _reflection.GeneratedProtocolMessageType('ResourceSlot', (_message.Message,), {
    'DESCRIPTOR' : _RESOURCEALLOCATIONS_RESOURCESLOT,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.ResourceAllocations.ResourceSlot)
    })
  ,
  'DESCRIPTOR' : _RESOURCEALLOCATIONS,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.ResourceAllocations)
  })
_sym_db.RegisterMessage(ResourceAllocations)
_sym_db.RegisterMessage(ResourceAllocations.ResourceSlot)

CoreWorkerStats = _reflection.GeneratedProtocolMessageType('CoreWorkerStats', (_message.Message,), {

  'UsedResourcesEntry' : _reflection.GeneratedProtocolMessageType('UsedResourcesEntry', (_message.Message,), {
    'DESCRIPTOR' : _COREWORKERSTATS_USEDRESOURCESENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.CoreWorkerStats.UsedResourcesEntry)
    })
  ,

  'WebuiDisplayEntry' : _reflection.GeneratedProtocolMessageType('WebuiDisplayEntry', (_message.Message,), {
    'DESCRIPTOR' : _COREWORKERSTATS_WEBUIDISPLAYENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.CoreWorkerStats.WebuiDisplayEntry)
    })
  ,
  'DESCRIPTOR' : _COREWORKERSTATS,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.CoreWorkerStats)
  })
_sym_db.RegisterMessage(CoreWorkerStats)
_sym_db.RegisterMessage(CoreWorkerStats.UsedResourcesEntry)
_sym_db.RegisterMessage(CoreWorkerStats.WebuiDisplayEntry)

NodeResourceUsage = _reflection.GeneratedProtocolMessageType('NodeResourceUsage', (_message.Message,), {
  'DESCRIPTOR' : _NODERESOURCEUSAGE,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NodeResourceUsage)
  })
_sym_db.RegisterMessage(NodeResourceUsage)

MetricPoint = _reflection.GeneratedProtocolMessageType('MetricPoint', (_message.Message,), {

  'TagsEntry' : _reflection.GeneratedProtocolMessageType('TagsEntry', (_message.Message,), {
    'DESCRIPTOR' : _METRICPOINT_TAGSENTRY,
    '__module__' : 'src.ray.protobuf.common_pb2'
    # @@protoc_insertion_point(class_scope:ray.rpc.MetricPoint.TagsEntry)
    })
  ,
  'DESCRIPTOR' : _METRICPOINT,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.MetricPoint)
  })
_sym_db.RegisterMessage(MetricPoint)
_sym_db.RegisterMessage(MetricPoint.TagsEntry)

NamedActorInfo = _reflection.GeneratedProtocolMessageType('NamedActorInfo', (_message.Message,), {
  'DESCRIPTOR' : _NAMEDACTORINFO,
  '__module__' : 'src.ray.protobuf.common_pb2'
  # @@protoc_insertion_point(class_scope:ray.rpc.NamedActorInfo)
  })
_sym_db.RegisterMessage(NamedActorInfo)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\030io.ray.runtime.generated\370\001\001'
  _JOBCONFIG_METADATAENTRY._options = None
  _JOBCONFIG_METADATAENTRY._serialized_options = b'8\001'
  _TASKSPEC_REQUIREDRESOURCESENTRY._options = None
  _TASKSPEC_REQUIREDRESOURCESENTRY._serialized_options = b'8\001'
  _TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY._options = None
  _TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY._serialized_options = b'8\001'
  _TASKINFOENTRY_REQUIREDRESOURCESENTRY._options = None
  _TASKINFOENTRY_REQUIREDRESOURCESENTRY._serialized_options = b'8\001'
  _BUNDLE_UNITRESOURCESENTRY._options = None
  _BUNDLE_UNITRESOURCESENTRY._serialized_options = b'8\001'
  _COREWORKERSTATS_USEDRESOURCESENTRY._options = None
  _COREWORKERSTATS_USEDRESOURCESENTRY._serialized_options = b'8\001'
  _COREWORKERSTATS_WEBUIDISPLAYENTRY._options = None
  _COREWORKERSTATS_WEBUIDISPLAYENTRY._serialized_options = b'8\001'
  _METRICPOINT_TAGSENTRY._options = None
  _METRICPOINT_TAGSENTRY._serialized_options = b'8\001'
  _LANGUAGE._serialized_start=11733
  _LANGUAGE._serialized_end=11774
  _WORKERTYPE._serialized_start=11776
  _WORKERTYPE._serialized_end=11850
  _TASKTYPE._serialized_start=11852
  _TASKTYPE._serialized_end=11937
  _ERRORTYPE._serialized_start=11940
  _ERRORTYPE._serialized_end=12584
  _TASKSTATUS._serialized_start=12587
  _TASKSTATUS._serialized_end=12831
  _WORKEREXITTYPE._serialized_start=12833
  _WORKEREXITTYPE._serialized_end=12957
  _PLACEMENTSTRATEGY._serialized_start=12959
  _PLACEMENTSTRATEGY._serialized_end=13036
  _NODEAFFINITYSCHEDULINGSTRATEGY._serialized_start=85
  _NODEAFFINITYSCHEDULINGSTRATEGY._serialized_end=162
  _PLACEMENTGROUPSCHEDULINGSTRATEGY._serialized_start=165
  _PLACEMENTGROUPSCHEDULINGSTRATEGY._serialized_end=388
  _DEFAULTSCHEDULINGSTRATEGY._serialized_start=390
  _DEFAULTSCHEDULINGSTRATEGY._serialized_end=417
  _SPREADSCHEDULINGSTRATEGY._serialized_start=419
  _SPREADSCHEDULINGSTRATEGY._serialized_end=445
  _SCHEDULINGSTRATEGY._serialized_start=448
  _SCHEDULINGSTRATEGY._serialized_end=934
  _ADDRESS._serialized_start=936
  _ADDRESS._serialized_end=1054
  _JAVAFUNCTIONDESCRIPTOR._serialized_start=1056
  _JAVAFUNCTIONDESCRIPTOR._serialized_end=1178
  _PYTHONFUNCTIONDESCRIPTOR._serialized_start=1181
  _PYTHONFUNCTIONDESCRIPTOR._serialized_end=1345
  _CPPFUNCTIONDESCRIPTOR._serialized_start=1347
  _CPPFUNCTIONDESCRIPTOR._serialized_end=1462
  _FUNCTIONDESCRIPTOR._serialized_start=1465
  _FUNCTIONDESCRIPTOR._serialized_end=1790
  _CONCURRENCYGROUP._serialized_start=1793
  _CONCURRENCYGROUP._serialized_end=1952
  _RAYERRORINFO._serialized_start=1955
  _RAYERRORINFO._serialized_end=2262
  _OUTOFMEMORYERRORCONTEXT._serialized_start=2265
  _OUTOFMEMORYERRORCONTEXT._serialized_end=2569
  _NODEDIEDERRORCONTEXT._serialized_start=2571
  _NODEDIEDERRORCONTEXT._serialized_end=2658
  _RAYEXCEPTION._serialized_start=2661
  _RAYEXCEPTION._serialized_end=2835
  _ACTORDEATHCAUSE._serialized_start=2838
  _ACTORDEATHCAUSE._serialized_end=3304
  _RUNTIMEENVFAILEDCONTEXT._serialized_start=3306
  _RUNTIMEENVFAILEDCONTEXT._serialized_end=3368
  _ACTORUNSCHEDULABLECONTEXT._serialized_start=3370
  _ACTORUNSCHEDULABLECONTEXT._serialized_end=3434
  _ACTORDIEDERRORCONTEXT._serialized_start=3437
  _ACTORDIEDERRORCONTEXT._serialized_end=3776
  _OOMCONTEXT._serialized_start=3778
  _OOMCONTEXT._serialized_end=3870
  _JOBCONFIG._serialized_start=3873
  _JOBCONFIG._serialized_end=4368
  _JOBCONFIG_METADATAENTRY._serialized_start=4260
  _JOBCONFIG_METADATAENTRY._serialized_end=4319
  _JOBCONFIG_ACTORLIFETIME._serialized_start=4321
  _JOBCONFIG_ACTORLIFETIME._serialized_end=4368
  _TASKSPEC._serialized_start=4371
  _TASKSPEC._serialized_end=6062
  _TASKSPEC_REQUIREDRESOURCESENTRY._serialized_start=5900
  _TASKSPEC_REQUIREDRESOURCESENTRY._serialized_end=5968
  _TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY._serialized_start=5970
  _TASKSPEC_REQUIREDPLACEMENTRESOURCESENTRY._serialized_end=6047
  _TASKINFOENTRY._serialized_start=6065
  _TASKINFOENTRY._serialized_end=6773
  _TASKINFOENTRY_REQUIREDRESOURCESENTRY._serialized_start=5900
  _TASKINFOENTRY_REQUIREDRESOURCESENTRY._serialized_end=5968
  _BUNDLE._serialized_start=6776
  _BUNDLE._serialized_end=7114
  _BUNDLE_BUNDLEIDENTIFIER._serialized_start=6949
  _BUNDLE_BUNDLEIDENTIFIER._serialized_end=7048
  _BUNDLE_UNITRESOURCESENTRY._serialized_start=7050
  _BUNDLE_UNITRESOURCESENTRY._serialized_end=7114
  _PLACEMENTGROUPSPEC._serialized_start=7117
  _PLACEMENTGROUPSPEC._serialized_end=7561
  _OBJECTREFERENCE._serialized_start=7564
  _OBJECTREFERENCE._serialized_end=7694
  _OBJECTREFERENCECOUNT._serialized_start=7697
  _OBJECTREFERENCECOUNT._serialized_end=8016
  _TASKARG._serialized_start=8019
  _TASKARG._serialized_end=8207
  _ACTORCREATIONTASKSPEC._serialized_start=8210
  _ACTORCREATIONTASKSPEC._serialized_end=8826
  _ACTORTASKSPEC._serialized_start=8829
  _ACTORTASKSPEC._serialized_end=8976
  _TASK._serialized_start=8978
  _TASK._serialized_end=9032
  _RESOURCEID._serialized_start=9034
  _RESOURCEID._serialized_end=9096
  _RESOURCEMAPENTRY._serialized_start=9098
  _RESOURCEMAPENTRY._serialized_end=9192
  _VIEWDATA._serialized_start=9195
  _VIEWDATA._serialized_end=9695
  _VIEWDATA_MEASURE._serialized_start=9292
  _VIEWDATA_MEASURE._serialized_end=9695
  _OBJECTREFINFO._serialized_start=9698
  _OBJECTREFINFO._serialized_end=10082
  _RESOURCEALLOCATIONS._serialized_start=10085
  _RESOURCEALLOCATIONS._serialized_end=10256
  _RESOURCEALLOCATIONS_RESOURCESLOT._serialized_start=10190
  _RESOURCEALLOCATIONS_RESOURCESLOT._serialized_end=10256
  _COREWORKERSTATS._serialized_start=10259
  _COREWORKERSTATS._serialized_end=11349
  _COREWORKERSTATS_USEDRESOURCESENTRY._serialized_start=11190
  _COREWORKERSTATS_USEDRESOURCESENTRY._serialized_end=11284
  _COREWORKERSTATS_WEBUIDISPLAYENTRY._serialized_start=11286
  _COREWORKERSTATS_WEBUIDISPLAYENTRY._serialized_end=11349
  _NODERESOURCEUSAGE._serialized_start=11351
  _NODERESOURCEUSAGE._serialized_end=11390
  _METRICPOINT._serialized_start=11393
  _METRICPOINT._serialized_end=11656
  _METRICPOINT_TAGSENTRY._serialized_start=11601
  _METRICPOINT_TAGSENTRY._serialized_end=11656
  _NAMEDACTORINFO._serialized_start=11658
  _NAMEDACTORINFO._serialized_end=11731
# @@protoc_insertion_point(module_scope)



import json
import logging
import types

from ray import cloudpickle as cloudpickle
from ray._private.utils import binary_to_hex, hex_to_binary
from ray.util.annotations import DeveloperAPI
from ray.util.debug import log_once

logger = logging.getLogger(__name__)


@DeveloperAPI
class TuneFunctionEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, types.FunctionType):
            return self._to_cloudpickle(obj)
        try:
            return super(TuneFunctionEncoder, self).default(obj)
        except Exception:
            if log_once(f"tune_func_encode:{str(obj)}"):
                logger.debug("Unable to encode. Falling back to cloudpickle.")
            return self._to_cloudpickle(obj)

    def _to_cloudpickle(self, obj):
        return {
            "_type": "CLOUDPICKLE_FALLBACK",
            "value": binary_to_hex(cloudpickle.dumps(obj)),
        }


@DeveloperAPI
class TuneFunctionDecoder(json.JSONDecoder):
    def __init__(self, *args, **kwargs):
        json.JSONDecoder.__init__(self, object_hook=self.object_hook, *args, **kwargs)

    def object_hook(self, obj):
        if obj.get("_type") == "CLOUDPICKLE_FALLBACK":
            return self._from_cloudpickle(obj)
        return obj

    def _from_cloudpickle(self, obj):
        return cloudpickle.loads(hex_to_binary(obj["value"]))



import argparse
from dataclasses import dataclass
from enum import Enum
import os.path
import tempfile
import typer
from typing import Optional
import requests


from ray.tune.experiment.config_parser import _make_parser
from ray.tune.result import DEFAULT_RESULTS_DIR


class FrameworkEnum(str, Enum):
    """Supported frameworks for RLlib, used for CLI argument validation."""

    tf = "tf"
    tf2 = "tf2"
    torch = "torch"


class SupportedFileType(str, Enum):
    """Supported file types for RLlib, used for CLI argument validation."""

    yaml = "yaml"
    python = "python"


def get_file_type(config_file: str) -> SupportedFileType:
    if config_file.endswith(".py"):
        file_type = SupportedFileType.python
    elif config_file.endswith(".yaml") or config_file.endswith(".yml"):
        file_type = SupportedFileType.yaml
    else:
        raise ValueError(
            "Unknown file type for config "
            "file: {}. Supported extensions: .py, "
            ".yml, .yaml".format(config_file)
        )
    return file_type


def _create_tune_parser_help():
    """Create a Tune dummy parser to access its 'help' docstrings."""
    parser = _make_parser(
        parser_creator=None,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    return parser.__dict__.get("_option_string_actions")


PARSER_HELP = _create_tune_parser_help()


def download_example_file(
    example_file: str,
    base_url: Optional[str] = "https://raw.githubusercontent.com/"
    + "ray-project/ray/master/rllib/",
):
    """Download the example file (e.g. from GitHub) if it doesn't exist locally.
    If the provided example file exists locally, we return it directly.

    Not every user will have cloned our repo and cd'ed into this working directory
    when using the CLI.

    Args:
        example_file: The example file to download.
        base_url: The base URL to download the example file from. Use this if
            'example_file' is a link relative to this base URL. If set to 'None',
            'example_file' is assumed to be a complete URL (or a local file, in which
            case nothing is downloaded).
    """
    temp_file = None
    if not os.path.exists(example_file):

        example_url = base_url + example_file if base_url else example_file
        print(f">>> Attempting to download example file {example_url}...")

        file_type = get_file_type(example_url)
        if file_type == SupportedFileType.yaml:
            temp_file = tempfile.NamedTemporaryFile(suffix=".yaml")
        else:
            assert (
                file_type == SupportedFileType.python
            ), f"`example_url` ({example_url}) must be a python or yaml file!"
            temp_file = tempfile.NamedTemporaryFile(suffix=".py")

        r = requests.get(example_url)
        with open(temp_file.name, "wb") as f:
            print(r.content)
            f.write(r.content)

        print(f"  Status code: {r.status_code}")
        if r.status_code == 200:
            print(f"  Downloaded example file to {temp_file.name}")
            # only overwrite the file if the download was successful
            example_file = temp_file.name

    return example_file, temp_file


def get_help(key: str) -> str:
    """Get the help string from a parser for a given key.
    If e.g. 'resource_group' is provided, we return
    the entry for '--resource-group'."""
    key = "--" + key
    key = key.replace("_", "-")
    if key not in PARSER_HELP.keys():
        raise ValueError(f"Key {key} not found in parser.")
    return PARSER_HELP.get(key).help


example_help = dict(
    filter="Filter examples by exact substring match. For instance,"
    " --filter=ppo will only show examples that"
    " contain the substring 'ppo' in their ID. The same way, -f=recsys"
    "will return all recommender system examples.",
)


train_help = dict(
    env="The environment specifier to use. This could be an Farama-Foundation "
    "Gymnasium specifier (e.g. `CartPole-v1`) or a full class-path (e.g. "
    "`ray.rllib.examples.env.simple_corridor.SimpleCorridor`).",
    config_file="Use the algorithm configuration from this file.",
    filetype="The file type of the config file. Defaults to 'yaml' and can also be "
    "'python'.",
    experiment_name="Name of the subdirectory under `local_dir` to put results in.",
    framework="The identifier of the deep learning framework you want to use."
    "Choose between TensorFlow 1.x ('tf'), TensorFlow 2.x ('tf2'), "
    "and PyTorch ('torch').",
    v="Whether to use INFO level logging.",
    vv="Whether to use DEBUG level logging.",
    resume="Whether to attempt to resume from previous experiments.",
    local_dir=f"Local dir to save training results to. "
    f"Defaults to '{DEFAULT_RESULTS_DIR}'.",
    local_mode="Run Ray in local mode for easier debugging.",
    ray_address="Connect to an existing Ray cluster at this address instead "
    "of starting a new one.",
    ray_ui="Whether to enable the Ray web UI.",
    ray_num_cpus="The '--num-cpus' argument to use if starting a new cluster.",
    ray_num_gpus="The '--num-gpus' argument to use if starting a new cluster.",
    ray_num_nodes="Emulate multiple cluster nodes for debugging.",
    ray_object_store_memory="--object-store-memory to use if starting a new cluster.",
    upload_dir="Optional URI to sync training results to (e.g. s3://bucket).",
    trace="Whether to attempt to enable tracing for eager mode.",
    torch="Whether to use PyTorch (instead of tf) as the DL framework. "
    "This argument is deprecated, please use --framework to select 'torch'"
    "as backend.",
    eager="Whether to attempt to enable TensorFlow eager execution. "
    "This argument is deprecated, please choose 'tf2' in "
    "--framework to run select eager mode.",
)


eval_help = dict(
    checkpoint="Optional checkpoint from which to roll out. If none provided, we will "
    "evaluate an untrained algorithm.",
    algo="The algorithm or model to train. This may refer to the name of a built-in "
    "Algorithm (e.g. RLlib's `DQN` or `PPO`), or a user-defined trainable "
    "function or class registered in the Tune registry.",
    env="The environment specifier to use. This could be an Farama-Foundation gymnasium"
    " specifier (e.g. `CartPole-v1`) or a full class-path (e.g. "
    "`ray.rllib.examples.env.simple_corridor.SimpleCorridor`).",
    local_mode="Run Ray in local mode for easier debugging.",
    render="Render the environment while evaluating. Off by default",
    video_dir="Specifies the directory into which videos of all episode"
    "rollouts will be stored.",
    steps="Number of time-steps to roll out. The evaluation will also stop if "
    "`--episodes` limit is reached first. A value of 0 means no "
    "limitation on the number of time-steps run.",
    episodes="Number of complete episodes to roll out. The evaluation will also stop "
    "if `--steps` (time-steps) limit is reached first. A value of 0 means "
    "no limitation on the number of episodes run.",
    out="Output filename",
    config="Algorithm-specific configuration (e.g. `env`, `framework` etc.). "
    "Gets merged with loaded configuration from checkpoint file and "
    "`evaluation_config` settings therein.",
    save_info="Save the info field generated by the step() method, "
    "as well as the action, observations, rewards and done fields.",
    use_shelve="Save rollouts into a Python shelf file (will save each episode "
    "as it is generated). An output filename must be set using --out.",
    track_progress="Write progress to a temporary file (updated "
    "after each episode). An output filename must be set using --out; "
    "the progress file will live in the same folder.",
)


@dataclass
class CLIArguments:
    """Dataclass for CLI arguments and options. We use this class to keep track
    of common arguments, like "run" or "env" that would otherwise be duplicated."""

    # Common arguments
    # __cli_common_start__
    Algo = typer.Option(None, "--algo", "--run", "-a", "-r", help=get_help("run"))
    AlgoRequired = typer.Option(
        ..., "--algo", "--run", "-a", "-r", help=get_help("run")
    )
    Env = typer.Option(None, "--env", "-e", help=train_help.get("env"))
    EnvRequired = typer.Option(..., "--env", "-e", help=train_help.get("env"))
    Config = typer.Option("{}", "--config", "-c", help=get_help("config"))
    ConfigRequired = typer.Option(..., "--config", "-c", help=get_help("config"))
    # __cli_common_end__

    # Train file arguments
    # __cli_file_start__
    ConfigFile = typer.Argument(  # config file is now mandatory for "file" subcommand
        ..., help=train_help.get("config_file")
    )
    FileType = typer.Option(
        SupportedFileType.yaml, "--type", "-t", help=train_help.get("filetype")
    )
    # __cli_file_end__

    # Train arguments
    # __cli_train_start__
    Stop = typer.Option("{}", "--stop", "-s", help=get_help("stop"))
    ExperimentName = typer.Option(
        "default", "--experiment-name", "-n", help=train_help.get("experiment_name")
    )
    V = typer.Option(False, "--log-info", "-v", help=train_help.get("v"))
    VV = typer.Option(False, "--log-debug", "-vv", help=train_help.get("vv"))
    Resume = typer.Option(False, help=train_help.get("resume"))
    NumSamples = typer.Option(1, help=get_help("num_samples"))
    CheckpointFreq = typer.Option(0, help=get_help("checkpoint_freq"))
    CheckpointAtEnd = typer.Option(True, help=get_help("checkpoint_at_end"))
    LocalDir = typer.Option(DEFAULT_RESULTS_DIR, help=train_help.get("local_dir"))
    Restore = typer.Option(None, help=get_help("restore"))
    Framework = typer.Option(None, help=train_help.get("framework"))
    ResourcesPerTrial = typer.Option(None, help=get_help("resources_per_trial"))
    KeepCheckpointsNum = typer.Option(None, help=get_help("keep_checkpoints_num"))
    CheckpointScoreAttr = typer.Option(
        "training_iteration", help=get_help("sync_on_checkpoint")
    )
    UploadDir = typer.Option("", help=train_help.get("upload_dir"))
    Trace = typer.Option(False, help=train_help.get("trace"))
    LocalMode = typer.Option(False, help=train_help.get("local_mode"))
    Scheduler = typer.Option("FIFO", help=get_help("scheduler"))
    SchedulerConfig = typer.Option("{}", help=get_help("scheduler_config"))
    RayAddress = typer.Option(None, help=train_help.get("ray_address"))
    RayUi = typer.Option(False, help=train_help.get("ray_ui"))
    RayNumCpus = typer.Option(None, help=train_help.get("ray_num_cpus"))
    RayNumGpus = typer.Option(None, help=train_help.get("ray_num_gpus"))
    RayNumNodes = typer.Option(None, help=train_help.get("ray_num_nodes"))
    RayObjectStoreMemory = typer.Option(
        None, help=train_help.get("ray_object_store_memory")
    )
    # __cli_train_end__

    # Eval arguments
    # __cli_eval_start__
    Checkpoint = typer.Argument(None, help=eval_help.get("checkpoint"))
    Render = typer.Option(False, help=eval_help.get("render"))
    Steps = typer.Option(10000, help=eval_help.get("steps"))
    Episodes = typer.Option(0, help=eval_help.get("episodes"))
    Out = typer.Option(None, help=eval_help.get("out"))
    SaveInfo = typer.Option(False, help=eval_help.get("save_info"))
    UseShelve = typer.Option(False, help=eval_help.get("use_shelve"))
    TrackProgress = typer.Option(False, help=eval_help.get("track_progress"))
    # __cli_eval_end__


# Note that the IDs of these examples are lexicographically sorted by environment,
# not by algorithm. This should be more natural for users, but could be changed easily.
EXAMPLES = {
    # A2C
    "atari-a2c": {
        "file": "tuned_examples/a2c/atari-a2c.yaml",
        "description": "Runs grid search over several Atari games on A2C.",
    },
    "cartpole-a2c": {
        "file": "tuned_examples/a2c/cartpole_a2c.py",
        "stop": "{'timesteps_total': 50000, 'episode_reward_mean': 200}",
        "description": "Runs A2C on the CartPole-v1 environment.",
    },
    "cartpole-a2c-micro": {
        "file": "tuned_examples/a2c/cartpole-a2c-microbatch.yaml",
        "description": "Runs A2C on the CartPole-v1 environment, using micro-batches.",
    },
    # A3C
    "cartpole-a3c": {
        "file": "tuned_examples/a3c/cartpole_a3c.py",
        "stop": "{'timesteps_total': 20000, 'episode_reward_mean': 150}",
        "description": "Runs A3C on the CartPole-v1 environment.",
    },
    "pong-a3c": {
        "file": "tuned_examples/a3c/pong-a3c.yaml",
        "description": "Runs A3C on the ALE/Pong-v5 (deterministic) environment.",
    },
    # AlphaStar
    "multi-agent-cartpole-alpha-star": {
        "file": "tuned_examples/alpha_star/multi-agent-cartpole-alpha-star.yaml",
        "description": "Runs AlphaStar on 4 CartPole agents.",
    },
    # AlphaZero
    "cartpole-alpha-zero": {
        "file": "tuned_examples/alpha_zero/cartpole-sparse-rewards-alpha-zero.yaml",
        "description": "Runs AlphaZero on a Cartpole with sparse rewards.",
    },
    # Apex DDPG
    "mountaincar-apex-ddpg": {
        "file": "tuned_examples/apex_ddpg/mountaincarcontinuous-apex-ddpg.yaml",
        "description": "Runs Apex DDPG on MountainCarContinuous-v0.",
    },
    "pendulum-apex-ddpg": {
        "file": "tuned_examples/apex_ddpg/pendulum-apex-ddpg.yaml",
        "description": "Runs Apex DDPG on Pendulum-v1.",
    },
    # Apex DQN
    "breakout-apex-dqn": {
        "file": "tuned_examples/apex_dqn/atari-apex-dqn.yaml",
        "description": "Runs Apex DQN on ALE/Breakout-v5 (no frameskip).",
    },
    "cartpole-apex-dqn": {
        "file": "tuned_examples/apex_dqn/cartpole-apex-dqn.yaml",
        "description": "Runs Apex DQN on CartPole-v1.",
    },
    "pong-apex-dqn": {
        "file": "tuned_examples/apex_dqn/pong-apex-dqn.yaml",
        "description": "Runs Apex DQN on ALE/Pong-v5 (no frameskip).",
    },
    # APPO
    "cartpole-appo": {
        "file": "tuned_examples/appo/cartpole-appo.yaml",
        "description": "Runs APPO on CartPole-v1.",
    },
    "frozenlake-appo": {
        "file": "tuned_examples/appo/frozenlake-appo-vtrace.yaml",
        "description": "Runs APPO on FrozenLake-v1.",
    },
    "halfcheetah-appo": {
        "file": "tuned_examples/appo/halfcheetah-appo.yaml",
        "description": "Runs APPO on HalfCheetah-v2.",
    },
    "multi-agent-cartpole-appo": {
        "file": "tuned_examples/appo/multi-agent-cartpole-appo.yaml",
        "description": "Runs APPO on RLlib's MultiAgentCartPole",
    },
    "pendulum-appo": {
        "file": "tuned_examples/appo/pendulum-appo.yaml",
        "description": "Runs APPO on Pendulum-v1.",
    },
    "pong-appo": {
        "file": "tuned_examples/appo/pong-appo.yaml",
        "description": "Runs APPO on ALE/Pong-v5 (no frameskip).",
    },
    # ARS
    "cartpole-ars": {
        "file": "tuned_examples/ars/cartpole-ars.yaml",
        "description": "Runs ARS on CartPole-v1.",
    },
    "swimmer-ars": {
        "file": "tuned_examples/ars/swimmer-ars.yaml",
        "description": "Runs ARS on Swimmer-v2.",
    },
    # Bandits
    "recsys-bandits": {
        "file": "tuned_examples/bandits/"
        + "interest-evolution-recsim-env-bandit-linucb.yaml",
        "description": "Runs BanditLinUCB on a Recommendation Simulation environment.",
    },
    # BC
    "cartpole-bc": {
        "file": "tuned_examples/bc/cartpole-bc.yaml",
        "description": "Runs BC on CartPole-v1.",
    },
    # CQL
    "halfcheetah-cql": {
        "file": "tuned_examples/cql/halfcheetah-cql.yaml",
        "description": "Runs grid search on HalfCheetah environments with CQL.",
    },
    "hopper-cql": {
        "file": "tuned_examples/cql/hopper-cql.yaml",
        "description": "Runs grid search on Hopper environments with CQL.",
    },
    "pendulum-cql": {
        "file": "tuned_examples/cql/pendulum-cql.yaml",
        "description": "Runs CQL on Pendulum-v1.",
    },
    # CRR
    "cartpole-crr": {
        "file": "tuned_examples/crr/CartPole-v1-crr.yaml",
        "description": "Run CRR on CartPole-v1.",
    },
    "pendulum-crr": {
        "file": "tuned_examples/crr/pendulum-v1-crr.yaml",
        "description": "Run CRR on Pendulum-v1.",
    },
    # DDPG
    "halfcheetah-ddpg": {
        "file": "tuned_examples/ddpg/halfcheetah-ddpg.yaml",
        "description": "Runs DDPG on HalfCheetah-v2.",
    },
    "halfcheetah-bullet-ddpg": {
        "file": "tuned_examples/ddpg/halfcheetah-pybullet-ddpg.yaml",
        "description": "Runs DDPG on HalfCheetahBulletEnv-v0.",
    },
    "hopper-bullet-ddpg": {
        "file": "tuned_examples/ddpg/hopper-pybullet-ddpg.yaml",
        "description": "Runs DDPG on HopperBulletEnv-v0.",
    },
    "mountaincar-ddpg": {
        "file": "tuned_examples/ddpg/mountaincarcontinuous-ddpg.yaml",
        "description": "Runs DDPG on MountainCarContinuous-v0.",
    },
    "pendulum-ddpg": {
        "file": "tuned_examples/ddpg/pendulum-ddpg.yaml",
        "description": "Runs DDPG on Pendulum-v1.",
    },
    # DDPPO
    "breakout-ddppo": {
        "file": "tuned_examples/ddppo/atari-ddppo.yaml",
        "description": "Runs DDPPO on ALE/Breakout-v5 (no frameskip).",
    },
    "cartpole-ddppo": {
        "file": "tuned_examples/ddppo/cartpole-ddppo.yaml",
        "description": "Runs DDPPO on CartPole-v1",
    },
    "pendulum-ddppo": {
        "file": "tuned_examples/ddppo/pendulum-ddppo.yaml",
        "description": "Runs DDPPO on Pendulum-v1.",
    },
    # DQN
    "atari-dqn": {
        "file": "tuned_examples/dqn/atari-dqn.yaml",
        "description": "Run grid search on Atari environments with DQN.",
    },
    "atari-duel-ddqn": {
        "file": "tuned_examples/dqn/atari-duel-ddqn.yaml",
        "description": "Run grid search on Atari environments "
        "with duelling double DQN.",
    },
    "cartpole-dqn": {
        "file": "tuned_examples/dqn/cartpole-dqn.yaml",
        "description": "Run DQN on CartPole-v1.",
    },
    "pong-dqn": {
        "file": "tuned_examples/dqn/pong-dqn.yaml",
        "description": "Run DQN on ALE/Pong-v5 (deterministic).",
    },
    "pong-rainbow": {
        "file": "tuned_examples/dqn/pong-rainbow.yaml",
        "description": "Run Rainbow on ALE/Pong-v5 (deterministic).",
    },
    # DREAMER
    "dm-control-dreamer": {
        "file": "tuned_examples/dreamer/dreamer-deepmind-control.yaml",
        "description": "Run DREAMER on a suite of control problems by Deepmind.",
    },
    # DT
    "cartpole-dt": {
        "file": "tuned_examples/dt/CartPole-v1-dt.yaml",
        "description": "Run DT on CartPole-v1.",
    },
    "pendulum-dt": {
        "file": "tuned_examples/dt/pendulum-v1-dt.yaml",
        "description": "Run DT on Pendulum-v1.",
    },
    # ES
    "cartpole-es": {
        "file": "tuned_examples/es/cartpole-es.yaml",
        "description": "Run ES on CartPole-v1.",
    },
    "humanoid-es": {
        "file": "tuned_examples/es/humanoid-es.yaml",
        "description": "Run ES on Humanoid-v2.",
    },
    # IMPALA
    "atari-impala": {
        "file": "tuned_examples/impala/atari-impala.yaml",
        "description": "Run grid search over several atari games with IMPALA.",
    },
    "cartpole-impala": {
        "file": "tuned_examples/impala/cartpole-impala.yaml",
        "description": "Run IMPALA on CartPole-v1.",
    },
    "multi-agent-cartpole-impala": {
        "file": "tuned_examples/impala/multi-agent-cartpole-impala.yaml",
        "description": "Run IMPALA on RLlib's MultiAgentCartPole",
    },
    "pendulum-impala": {
        "file": "tuned_examples/impala/pendulum-impala.yaml",
        "description": "Run IMPALA on Pendulum-v1.",
    },
    "pong-impala": {
        "file": "tuned_examples/impala/pong-impala-fast.yaml",
        "description": "Run IMPALA on ALE/Pong-v5 (no frameskip).",
    },
    # MADDPG
    "two-step-game-maddpg": {
        "file": "tuned_examples/maddpg/two-step-game-maddpg.yaml",
        "description": "Run RLlib's Two-step game with multi-agent DDPG.",
    },
    # MAML
    "cartpole-maml": {
        "file": "tuned_examples/maml/cartpole-maml.yaml",
        "description": "Run MAML on CartPole-v1.",
    },
    "halfcheetah-maml": {
        "file": "tuned_examples/maml/halfcheetah-rand-direc-maml.yaml",
        "description": "Run MAML on a custom HalfCheetah environment.",
    },
    "pendulum-maml": {
        "file": "tuned_examples/maml/pendulum-mass-maml.yaml",
        "description": "Run MAML on a custom Pendulum environment.",
    },
    # MARWIL
    "cartpole-marwil": {
        "file": "tuned_examples/marwil/cartpole-marwil.yaml",
        "description": "Run MARWIL on CartPole-v1.",
    },
    # MBMPO
    "cartpole-mbmpo": {
        "file": "tuned_examples/mbmpo/cartpole-mbmpo.yaml",
        "description": "Run MBMPO on a CartPole environment wrapper.",
    },
    "halfcheetah-mbmpo": {
        "file": "tuned_examples/mbmpo/halfcheetah-mbmpo.yaml",
        "description": "Run MBMPO on a HalfCheetah environment wrapper.",
    },
    "hopper-mbmpo": {
        "file": "tuned_examples/mbmpo/hopper-mbmpo.yaml",
        "description": "Run MBMPO on a Hopper environment wrapper.",
    },
    "pendulum-mbmpo": {
        "file": "tuned_examples/mbmpo/pendulum-mbmpo.yaml",
        "description": "Run MBMPO on a Pendulum environment wrapper.",
    },
    # PG
    "cartpole-pg": {
        "file": "tuned_examples/pg/cartpole-pg.yaml",
        "description": "Run PG on CartPole-v1",
    },
    # PPO
    "atari-ppo": {
        "file": "tuned_examples/ppo/atari-ppo.yaml",
        "description": "Run grid search over several atari games with PPO.",
    },
    "cartpole-ppo": {
        "file": "tuned_examples/ppo/cartpole-ppo.yaml",
        "description": "Run PPO on CartPole-v1.",
    },
    "halfcheetah-ppo": {
        "file": "tuned_examples/ppo/halfcheetah-ppo.yaml",
        "description": "Run PPO on HalfCheetah-v2.",
    },
    "hopper-ppo": {
        "file": "tuned_examples/ppo/hopper-ppo.yaml",
        "description": "Run PPO on Hopper-v1.",
    },
    "humanoid-ppo": {
        "file": "tuned_examples/ppo/humanoid-ppo.yaml",
        "description": "Run PPO on Humanoid-v1.",
    },
    "pendulum-ppo": {
        "file": "tuned_examples/ppo/pendulum-ppo.yaml",
        "description": "Run PPO on Pendulum-v1.",
    },
    "pong-ppo": {
        "file": "tuned_examples/ppo/pong-ppo.yaml",
        "description": "Run PPO on ALE/Pong-v5 (no frameskip).",
    },
    "recsys-ppo": {
        "file": "tuned_examples/ppo/recomm-sys001-ppo.yaml",
        "description": "Run PPO on a recommender system example from RLlib.",
    },
    "repeatafterme-ppo": {
        "file": "tuned_examples/ppo/repeatafterme-ppo-lstm.yaml",
        "description": "Run PPO on RLlib's RepeatAfterMe environment.",
    },
    "walker2d-ppo": {
        "file": "tuned_examples/ppo/walker2d-ppo.yaml",
        "description": "Run PPO on the Walker2d-v1 environment.",
    },
    # QMIX
    "two-step-game-qmix": {
        "file": "tuned_examples/qmix/two-step-game-qmix.yaml",
        "description": "Run QMIX on RLlib's two-step game.",
    },
    # R2D2
    "stateless-cartpole-r2d2": {
        "file": "tuned_examples/r2d2/stateless-cartpole-r2d2.yaml",
        "description": "Run R2D2 on a stateless cart pole environment.",
    },
    # SAC
    "atari-sac": {
        "file": "tuned_examples/sac/atari-sac.yaml",
        "description": "Run grid search on several atari games with SAC.",
    },
    "cartpole-sac": {
        "file": "tuned_examples/sac/cartpole-sac.yaml",
        "description": "Run SAC on CartPole-v1",
    },
    "halfcheetah-sac": {
        "file": "tuned_examples/sac/halfcheetah-sac.yaml",
        "description": "Run SAC on HalfCheetah-v3.",
    },
    "pacman-sac": {
        "file": "tuned_examples/sac/mspacman-sac.yaml",
        "description": "Run SAC on ALE/MsPacman-v5 (no frameskip).",
    },
    "pendulum-sac": {
        "file": "tuned_examples/sac/pendulum-sac.yaml",
        "description": "Run SAC on Pendulum-v1.",
    },
    # SimpleQ
    "cartpole-simpleq": {
        "file": "tuned_examples/simple_q/cartpole-simpleq.yaml",
        "description": "Run SimpleQ on CartPole-v1",
    },
    # SlateQ
    "recsys-long-term-slateq": {
        "file": "tuned_examples/slateq/long-term-satisfaction-recsim-env-slateq.yaml",
        "description": "Run SlateQ on a recommendation system aimed at "
        "long-term satisfaction.",
    },
    "recsys-parametric-slateq": {
        "file": "tuned_examples/slateq/parametric-item-reco-env-slateq.yaml",
        "description": "SlateQ run on a recommendation system.",
    },
    "recsys-slateq": {
        "file": "tuned_examples/slateq/recomm-sys001-slateq.yaml",
        "description": "SlateQ run on a recommendation system.",
    },
    # TD3
    "inverted-pendulum-td3": {
        "file": "tuned_examples/td3/invertedpendulum-td3.yaml",
        "description": "Run TD3 on InvertedPendulum-v2.",
    },
    "mujoco-td3": {
        "file": "tuned_examples/td3/mujoco-td3.yaml",
        "description": "Run TD3 against four of the hardest MuJoCo tasks.",
    },
    "pendulum-td3": {
        "file": "tuned_examples/td3/pendulum-td3.yaml",
        "description": "Run TD3 on Pendulum-v1.",
    },
}



import os
import tempfile
import time

import pytest
import sys
import io
from contextlib import redirect_stdout

import ray
from ray import tune
from ray.air import session, RunConfig
from ray.tune.progress_reporter import JupyterNotebookReporter
from ray.util.client.ray_client_helpers import ray_start_client_server


@pytest.fixture
def start_client_server():
    with ray_start_client_server() as client:
        yield client
    ray.shutdown()


@pytest.fixture
def start_client_server_2_cpus():
    ray.init(num_cpus=2)
    with ray_start_client_server() as client:
        yield client
    ray.shutdown()


@pytest.fixture
def start_client_server_4_cpus():
    ray.init(num_cpus=4)
    with ray_start_client_server() as client:
        yield client
    ray.shutdown()


def test_pbt_function(start_client_server_2_cpus):
    assert ray.util.client.ray.is_connected()
    from ray.tune.examples.pbt_function import run_tune_pbt

    run_tune_pbt()


def test_optuna_example(start_client_server):
    assert ray.util.client.ray.is_connected()
    from ray.tune.examples.optuna_example import run_optuna_tune

    run_optuna_tune(smoke_test=True)


def test_cifar10_pytorch(start_client_server_2_cpus):
    assert ray.util.client.ray.is_connected()
    from ray.tune.examples.cifar10_pytorch import main

    main(num_samples=1, max_num_epochs=1, gpus_per_trial=0)


def test_tune_mnist_keras(start_client_server_4_cpus):
    assert ray.util.client.ray.is_connected()
    from ray.tune.examples.tune_mnist_keras import tune_mnist

    tune_mnist(num_training_iterations=5)


def test_mnist_ptl_mini(start_client_server):
    assert ray.util.client.ray.is_connected()
    from ray.tune.examples.mnist_ptl_mini import tune_mnist

    tune_mnist(num_samples=1, num_epochs=1, gpus_per_trial=0)


def test_xgboost_example(start_client_server):
    assert ray.util.client.ray.is_connected()
    from ray.tune.examples.xgboost_example import tune_xgboost

    tune_xgboost()


def test_xgboost_dynamic_resources_example(start_client_server):
    assert ray.util.client.ray.is_connected()
    from ray.tune.examples.xgboost_dynamic_resources_example import tune_xgboost

    tune_xgboost(use_class_trainable=True)
    tune_xgboost(use_class_trainable=False)


def test_mlflow_example(start_client_server):
    assert ray.util.client.ray.is_connected()
    from ray.tune.examples.mlflow_example import tune_with_callback, tune_with_setup

    mlflow_tracking_uri = os.path.join(tempfile.gettempdir(), "mlruns")
    tune_with_callback(mlflow_tracking_uri, finish_fast=True)
    tune_with_setup(mlflow_tracking_uri, finish_fast=True)


def test_pbt_transformers(start_client_server):
    assert ray.util.client.ray.is_connected()
    from ray.tune.examples.pbt_transformers.pbt_transformers import tune_transformer

    tune_transformer(num_samples=1, gpus_per_trial=0, smoke_test=True)


def test_jupyter_rich_output(start_client_server_4_cpus):
    assert ray.util.client.ray.is_connected()

    def dummy_objective(config):
        time.sleep(1)
        session.report(dict(metric=1))

    ip = ray.util.get_node_ip_address()

    class MockJupyterNotebookReporter(JupyterNotebookReporter):
        def display(self, string: str) -> None:
            # Make sure display is called on the driver
            assert ip == ray.util.get_node_ip_address()
            if string:
                assert "<div" in string
                print(string)

    reporter = MockJupyterNotebookReporter()
    buffer = io.StringIO()
    with redirect_stdout(buffer):
        tune.run(dummy_objective, num_samples=2, progress_reporter=reporter)
        print("", flush=True)
        assert "<div" in buffer.getvalue()

    reporter = MockJupyterNotebookReporter()
    buffer = io.StringIO()
    with redirect_stdout(buffer):
        tuner = tune.Tuner(
            dummy_objective, run_config=RunConfig(progress_reporter=reporter)
        )
        tuner.fit()
        print("", flush=True)
        assert "<div" in buffer.getvalue()


if __name__ == "__main__":
    import pytest

    sys.exit(pytest.main(["-v", __file__]))



import logging
import gymnasium as gym

from ray.rllib.offline.input_reader import InputReader
from ray.rllib.offline.io_context import IOContext
from ray.rllib.policy.sample_batch import SampleBatch
from ray.rllib.utils.annotations import override, PublicAPI
from ray.rllib.utils.typing import SampleBatchType
from typing import Dict

logger = logging.getLogger(__name__)


@PublicAPI
class D4RLReader(InputReader):
    """Reader object that loads the dataset from the D4RL dataset."""

    @PublicAPI
    def __init__(self, inputs: str, ioctx: IOContext = None):
        """Initializes a D4RLReader instance.

        Args:
            inputs: String corresponding to the D4RL environment name.
            ioctx: Current IO context object.
        """
        import d4rl

        self.env = gym.make(inputs)
        self.dataset = _convert_to_batch(d4rl.qlearning_dataset(self.env))
        assert self.dataset.count >= 1
        self.counter = 0

    @override(InputReader)
    def next(self) -> SampleBatchType:
        if self.counter >= self.dataset.count:
            self.counter = 0

        self.counter += 1
        return self.dataset.slice(start=self.counter, end=self.counter + 1)


def _convert_to_batch(dataset: Dict) -> SampleBatchType:
    # Converts D4RL dataset to SampleBatch
    d = {}
    d[SampleBatch.OBS] = dataset["observations"]
    d[SampleBatch.ACTIONS] = dataset["actions"]
    d[SampleBatch.NEXT_OBS] = dataset["next_observations"]
    d[SampleBatch.REWARDS] = dataset["rewards"]
    d[SampleBatch.TERMINATEDS] = dataset["terminals"]

    return SampleBatch(d)



import logging
import numpy as np
import time

from ray.rllib.models.jax.jax_modelv2 import JAXModelV2
from ray.rllib.models.jax.misc import SlimFC
from ray.rllib.utils.annotations import override
from ray.rllib.utils.framework import try_import_jax

jax, flax = try_import_jax()

logger = logging.getLogger(__name__)


class FullyConnectedNetwork(JAXModelV2):
    """Generic fully connected network."""

    def __init__(self, obs_space, action_space, num_outputs, model_config, name):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)

        self.key = jax.random.PRNGKey(int(time.time()))

        activation = model_config.get("fcnet_activation")
        hiddens = model_config.get("fcnet_hiddens", [])
        no_final_linear = model_config.get("no_final_linear")
        self.vf_share_layers = model_config.get("vf_share_layers")
        self.free_log_std = model_config.get("free_log_std")

        # Generate free-floating bias variables for the second half of
        # the outputs.
        if self.free_log_std:
            assert num_outputs % 2 == 0, (
                "num_outputs must be divisible by two",
                num_outputs,
            )
            num_outputs = num_outputs // 2

        self._hidden_layers = []
        prev_layer_size = int(np.product(obs_space.shape))
        self._logits = None

        # Create layers 0 to second-last.
        for size in hiddens[:-1]:
            self._hidden_layers.append(
                SlimFC(in_size=prev_layer_size, out_size=size, activation_fn=activation)
            )
            prev_layer_size = size

        # The last layer is adjusted to be of size num_outputs, but it's a
        # layer with activation.
        if no_final_linear and num_outputs:
            self._hidden_layers.append(
                SlimFC(
                    in_size=prev_layer_size,
                    out_size=num_outputs,
                    activation_fn=activation,
                )
            )
            prev_layer_size = num_outputs
        # Finish the layers with the provided sizes (`hiddens`), plus -
        # iff num_outputs > 0 - a last linear layer of size num_outputs.
        else:
            if len(hiddens) > 0:
                self._hidden_layers.append(
                    SlimFC(
                        in_size=prev_layer_size,
                        out_size=hiddens[-1],
                        activation_fn=activation,
                    )
                )
                prev_layer_size = hiddens[-1]
            if num_outputs:
                self._logits = SlimFC(
                    in_size=prev_layer_size, out_size=num_outputs, activation_fn=None
                )
            else:
                self.num_outputs = ([int(np.product(obs_space.shape))] + hiddens[-1:])[
                    -1
                ]

        # Layer to add the log std vars to the state-dependent means.
        if self.free_log_std and self._logits:
            raise ValueError("`free_log_std` not supported for JAX yet!")

        self._value_branch_separate = None
        if not self.vf_share_layers:
            # Build a parallel set of hidden layers for the value net.
            prev_vf_layer_size = int(np.product(obs_space.shape))
            vf_layers = []
            for size in hiddens:
                vf_layers.append(
                    SlimFC(
                        in_size=prev_vf_layer_size,
                        out_size=size,
                        activation_fn=activation,
                    )
                )
                prev_vf_layer_size = size
            self._value_branch_separate = vf_layers

        self._value_branch = SlimFC(
            in_size=prev_layer_size, out_size=1, activation_fn=None
        )
        # Holds the current "base" output (before logits layer).
        self._features = None
        # Holds the last input, in case value branch is separate.
        self._last_flat_in = None

    @override(JAXModelV2)
    def forward(self, input_dict, state, seq_lens):
        self._last_flat_in = input_dict["obs_flat"]
        x = self._last_flat_in
        for layer in self._hidden_layers:
            x = layer(x)
        self._features = x
        logits = self._logits(self._features) if self._logits else self._features
        if self.free_log_std:
            logits = self._append_free_log_std(logits)
        return logits, state

    @override(JAXModelV2)
    def value_function(self):
        assert self._features is not None, "must call forward() first"
        if self._value_branch_separate:
            return self._value_branch(
                self._value_branch_separate(self._last_flat_in)
            ).squeeze(1)
        else:
            return self._value_branch(self._features).squeeze(1)



import copy
import numpy as np
import pandas as pd
from typing import Callable, Dict, Any

import ray
from ray.data import Dataset

from ray.rllib.policy import Policy
from ray.rllib.policy.sample_batch import SampleBatch, convert_ma_batch_to_sample_batch
from ray.rllib.utils.annotations import override, DeveloperAPI, ExperimentalAPI
from ray.rllib.utils.typing import SampleBatchType
from ray.rllib.offline.offline_evaluator import OfflineEvaluator


@DeveloperAPI
def _perturb_fn(batch: np.ndarray, index: int):
    # shuffle the indexth column features
    random_inds = np.random.permutation(batch.shape[0])
    batch[:, index] = batch[random_inds, index]


@ExperimentalAPI
def _perturb_df(batch: pd.DataFrame, index: int):
    obs_batch = np.vstack(batch["obs"].values)
    _perturb_fn(obs_batch, index)
    batch["perturbed_obs"] = list(obs_batch)
    return batch


def _compute_actions(
    batch: pd.DataFrame,
    policy_state: Dict[str, Any],
    input_key: str = "",
    output_key: str = "",
):
    """A custom local function to do batch prediction of a policy.

    Given the policy state the action predictions are computed as a function of
    `input_key` and stored in the `output_key` column.

    Args:
        batch: A sub-batch from the dataset.
        policy_state: The state of the policy to use for the prediction.
        input_key: The key to use for the input to the policy. If not given, the
            default is SampleBatch.OBS.
        output_key: The key to use for the output of the policy. If not given, the
            default is "predicted_actions".

    Returns:
        The modified batch with the predicted actions added as a column.
    """
    if not input_key:
        input_key = SampleBatch.OBS

    policy = Policy.from_state(policy_state)
    sample_batch = SampleBatch(
        {
            SampleBatch.OBS: np.vstack(batch[input_key].values),
        }
    )
    actions, _, _ = policy.compute_actions_from_input_dict(sample_batch, explore=False)

    if not output_key:
        output_key = "predicted_actions"
    batch[output_key] = actions

    return batch


@ray.remote
def get_feature_importance_on_index(
    dataset: ray.data.Dataset,
    *,
    index: int,
    perturb_fn: Callable[[pd.DataFrame, int], None],
    batch_size: int,
    policy_state: Dict[str, Any],
):
    """A remote function to compute the feature importance of a given index.

    Args:
        dataset: The dataset to use for the computation. The dataset should have `obs`
            and `actions` columns. Each record should be flat d-dimensional array.
        index: The index of the feature to compute the importance for.
        perturb_fn: The function to use for perturbing the dataset at the given index.
        batch_size: The batch size to use for the computation.
        policy_state: The state of the policy to use for the computation.

    Returns:
        The modified dataset that contains a `delta` column which is the absolute
        difference between the expected output and the output due to the perturbation.
    """
    perturbed_ds = dataset.map_batches(
        perturb_fn, batch_size=batch_size, fn_kwargs={"index": index}
    )
    perturbed_actions = perturbed_ds.map_batches(
        _compute_actions,
        batch_size=batch_size,
        fn_kwargs={
            "output_key": "perturbed_actions",
            "input_key": "perturbed_obs",
            "policy_state": policy_state,
        },
    )

    def delta_fn(batch):
        # take the abs difference between columns 'ref_actions` and `perturbed_actions`
        # and store it in `diff`
        batch["delta"] = np.abs(batch["ref_actions"] - batch["perturbed_actions"])
        return batch

    delta = perturbed_actions.map_batches(delta_fn, batch_size=batch_size)

    return delta


@DeveloperAPI
class FeatureImportance(OfflineEvaluator):
    @override(OfflineEvaluator)
    def __init__(
        self,
        policy: Policy,
        repeat: int = 1,
        limit_fraction: float = 1.0,
        perturb_fn: Callable[[pd.DataFrame, int], pd.DataFrame] = _perturb_df,
    ):
        """Feature importance in a model inspection technique that can be used for any
        fitted predictor when the data is tablular.

        This implementation is also known as permutation importance that is defined to
        be the variation of the model's prediction when a single feature value is
        randomly shuffled. In RLlib it is implemented as a custom OffPolicyEstimator
        which is used to evaluate RLlib policies without performing environment
        interactions.

        Example usage: In the example below the feature importance module is used to
        evaluate the policy and the each feature's importance is computed after each
        training iteration. The permutation are repeated `self.repeat` times and the
        results are averages across repeats.

        ```python
            config = (
                AlgorithmConfig()
                .offline_data(
                    off_policy_estimation_methods=
                        {
                            "feature_importance": {
                                "type": FeatureImportance,
                                "repeat": 10,
                                "limit_fraction": 0.1,
                            }
                        }
                )
            )

            algorithm = DQN(config=config)
            results = algorithm.train()
        ```

        Args:
            policy: the policy to use for feature importance.
            repeat: number of times to repeat the perturbation.
            perturb_fn: function to perturb the features. By default reshuffle the
                features within the batch.
            limit_fraction: fraction of the dataset to use for feature importance
                This is only used in estimate_on_dataset when the dataset is too large
                to compute feature importance on.
        """
        super().__init__(policy)
        self.repeat = repeat
        self.perturb_fn = perturb_fn
        self.limit_fraction = limit_fraction

    def estimate(self, batch: SampleBatchType) -> Dict[str, Any]:
        """Estimate the feature importance of the policy.

        Given a batch of tabular observations, the importance of each feature is
        computed by perturbing each feature and computing the difference between the
        perturbed policy and the reference policy. The importance is computed for each
        feature and each perturbation is repeated `self.repeat` times.

        Args:
            batch: the batch of data to use for feature importance.

        Returns:
            A dict mapping each feature index string to its importance.
        """
        batch = convert_ma_batch_to_sample_batch(batch)
        obs_batch = batch["obs"]
        n_features = obs_batch.shape[-1]
        importance = np.zeros((self.repeat, n_features))

        ref_actions, _, _ = self.policy.compute_actions(obs_batch, explore=False)
        for r in range(self.repeat):
            for i in range(n_features):
                copy_obs_batch = copy.deepcopy(obs_batch)
                _perturb_fn(copy_obs_batch, index=i)
                perturbed_actions, _, _ = self.policy.compute_actions(
                    copy_obs_batch, explore=False
                )
                importance[r, i] = np.mean(np.abs(perturbed_actions - ref_actions))

        # take an average across repeats
        importance = importance.mean(0)
        metrics = {f"feature_{i}": importance[i] for i in range(len(importance))}

        return metrics

    @override(OfflineEvaluator)
    def estimate_on_dataset(
        self, dataset: Dataset, *, n_parallelism: int = ...
    ) -> Dict[str, Any]:
        """Estimate the feature importance of the policy given a dataset.

        For each feature in the dataset, the importance is computed by applying
        perturbations to each feature and computing the difference between the
        perturbed prediction and the reference prediction. The importance
        computation for each feature and each perturbation is repeated `self.repeat`
        times. If dataset is large the user can initialize the estimator with a
        `limit_fraction` to limit the dataset to a fraction of the original dataset.

        The dataset should include a column named `obs` where each row is a vector of D
        dimensions. The importance is computed for each dimension of the vector.

        Note (Implementation detail): The computation across features are distributed
        with ray workers since each feature is independent of each other.

        Args:
            dataset: the dataset to use for feature importance.
            n_parallelism: number of parallel workers to use for feature importance.

        Returns:
            A dict mapping each feature index string to its importance.
        """

        policy_state = self.policy.get_state()
        # step 1: limit the dataset to a few first rows
        ds = dataset.limit(int(self.limit_fraction * dataset.count()))

        # step 2: compute the reference actions
        bsize = max(1, ds.count() // n_parallelism)
        actions_ds = ds.map_batches(
            _compute_actions,
            batch_size=bsize,
            fn_kwargs={
                "output_key": "ref_actions",
                "policy_state": policy_state,
            },
        )

        # step 3: compute the feature importance
        n_features = ds.take(1)[0][SampleBatch.OBS].shape[-1]
        importance = np.zeros((self.repeat, n_features))
        for r in range(self.repeat):
            # shuffle the entire dataset
            shuffled_ds = actions_ds.random_shuffle()
            bsize_per_task = max(1, (shuffled_ds.count() * n_features) // n_parallelism)

            # for each index perturb the dataset and compute the feat importance score
            remote_fns = [
                get_feature_importance_on_index.remote(
                    dataset=shuffled_ds,
                    index=i,
                    perturb_fn=self.perturb_fn,
                    bsize=bsize_per_task,
                    policy_state=policy_state,
                )
                for i in range(n_features)
            ]
            ds_w_fi_scores = ray.get(remote_fns)
            importance[r] = np.array([d.mean("delta") for d in ds_w_fi_scores])

        importance = importance.mean(0)
        metrics = {f"feature_{i}": importance[i] for i in range(len(importance))}

        return metrics
